'''Validation and parsing of text enclosed by opening and closing delimiters (i.e. parentheses-like behavior)'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Generator, Iterable, Union

import re
from functools import partial
from enum import StrEnum


# PARSING AND VALIDATION METHODS FOR GENERAL OPENING/CLOSING DELIMITERS
class DelimiterBalanceState(StrEnum):
    '''For indication the error state in a delimiter character pair closure check'''
    BALANCED = 'balanced'
    PREMATURE_CLOSURE = 'premature closure'
    UNCLOSED_OPENING  = 'unclosed opening'

def check_balanced_delimiters(string : str, start_char : str, end_char : str) -> tuple[bool, DelimiterBalanceState, int]:
    '''Tests if parenthesis-like start and end delimiters within a string are balanced
    (i.e. there are just as many closures as opening, and no closure occurs before and opening)
    
    Return whether the string is balanced, and the index where an imbalance has occurred (-1 if no imbalance or at end)'''
    stack : list[int] = []
    for i, char in enumerate(string):
        if char == start_char:
            stack.append(i)
        if char == end_char:
            try:
                stack.pop()
            except IndexError:
                return False, DelimiterBalanceState.PREMATURE_CLOSURE, i
    if stack:
        return False, DelimiterBalanceState.UNCLOSED_OPENING, stack.pop() # even if no premature closures are present, imbalance can still occur if too few closures have happened by end
    return True, DelimiterBalanceState.BALANCED, -1 # -1 indicates no error here

def parse_within_delimiters(string : str, start_char : str, end_char : str) -> Generator[str, None, None]:
    '''Generates substring contained by starting and ending delimiting characters (and depth of substring) via pushdown automaton mechanism'''
    stack : list[int] = []
    for i, char in enumerate(string):
        if char == start_char:
            stack.append(i)
        if (char == end_char) and stack:
            start_idx = stack.pop()
            yield string[start_idx+1:i], len(stack)


# DELIMITER INSERTION METHODS
def delimit_substrs_from_regex(string : str, regex : Union[str, re.Pattern], start_char : str, end_char : str) -> str:
    '''Place delimiters around any substrings of "string" which match the provided regular expression'''
    return re.sub(regex, rf"{start_char}\1{end_char}", string)

def delimit_substrs_from_choices(string : str, choices : Iterable[str], start_char : str, end_char : str) -> str:
    '''Place delimiters around any substrings of "string" which occur within "choices"'''
    choices_re = re.compile(f'({"|".join(choices)})') # choice strings OR'ed together and surrounded with capture group parens  TODO: check that no choices are contained within any other choices
    return delimit_substrs_from_regex(string, regex=choices_re, start_char=start_char, end_char=end_char)


# BOILERPLATE FUNCTIONALITY FOR COMMON PARENTHETICAL DELIMITERS
COMMON_DELIMITERS = {
    'parentheses'     : '()',
    'square_brackets' : '[]',
    'curly_brackets'  : '{}',
    'chevrons'        : '<>', 
}

for delimiter_name, (start_char, end_char) in COMMON_DELIMITERS.items(): # register specific cases as globally-accessible functions
    globals()[f'parse_{delimiter_name}'         ] = partial(parse_within_delimiters     , start_char=start_char, end_char=end_char)
    globals()[f'check_balanced_{delimiter_name}'] = partial(check_balanced_delimiters   , start_char=start_char, end_char=end_char)
    globals()[f'{delimiter_name}_around_letters'] = partial(delimit_substrs_from_regex  , start_char=start_char, end_char=end_char, regex=re.compile('([A-Za-z])'), )
    globals()[f'{delimiter_name}_around_choices'] = partial(delimit_substrs_from_choices, start_char=start_char, end_char=end_char)

def validate_common_delimiters(string : str, pointer_char : str='^') -> None:
    '''For checking whether all braces-like delimiters (i.e. parenthesis, brackets, etc.) are fully opened and not closed early
    If an invalid delimiters are found, a ValueError is raised with a helpful error message that points to the location of the fault (and the reason)'''
    for delimiter_name, (start_char, end_char) in COMMON_DELIMITERS.items():
        is_balanced, balance_state, err_idx = check_balanced_delimiters(string, start_char=start_char, end_char=end_char)
        if not is_balanced:
            indicator_str = err_idx*' ' + pointer_char
            raise ValueError(f'Imbalanced {delimiter_name} found in string (reason: {balance_state.value})\n{string}\n{indicator_str}')
validate_braces = validate_common_delimiters # alias for convenience


'''For computing statistical measures and series' of structured data'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import numpy as np
from dataclasses import dataclass


# CALCULATION OF STATISTICS
def RMSE(pred : np.ndarray, obs : np.ndarray) -> float:
    '''Computes root-mean squared error between predicted and observed sets of values'''
    sq_err = (obs - pred)**2
    return np.sqrt(sq_err.mean())

@dataclass
class Accumulator:
    '''Compact container for accumulating averages'''
    sum : float = 0.0
    count : int = 0

    @property
    def average(self) -> float:
        return self.sum / self.count
    
    
# SERIES-TO-SERIES OPERATIONS
def normalize(series : np.ndarray) -> np.ndarray:
    '''Normalize a series of data to have all values lie in the range [-1, 1]'''
    return (series - series.min()) / (series.max() - series.min())

def standardize(series : np.ndarray) -> np.ndarray:
    '''Standardize a series of data to have mean 0 and standard deviation of 1'''
    return (series - series.mean()) / series.std() 

def autocorrelate(series : np.ndarray) -> np.ndarray:
    '''Compute autocorrelation of a vector series of data points'''
    assert(series.ndim == 1) # only operate for vectors of data
    
    series = standardize(series)
    autocorr = np.correlate(series, series, mode='full') # need full mode to get proper length
    autocorr = autocorr[autocorr.size//2:] # only keep second half of array (Hermitian, so symmetric about midpoint)
    autocorr /= series.size # normalize to range [-1, 1] by including number of datapoints (due to variance estimator)

    return autocorr



'''For mapping 1-to-1 between two allegedly identical molecules'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Generator
from rdkit.Chem.rdchem import Atom, Mol

from ..rderrors import SubstructMatchFailedError


# CUSTOM EXCEPTIONS
class MolSizeMismatchError(Exception):
    '''Raised when an operation is attempted on two molecules which were expected to have the same size but don't'''
    pass

# BIJECTIVE MATCHING METHODS
def bijective_atom_id_iter(rdmol_1 : Mol, rdmol_2 : Mol) -> Generator[tuple[int, int], None, None]:
    '''Takes two chemically identical molecules, matches corresponding atoms between them 1:1, and generates matching atom id pairs
    Yields atoms in pairs in the same order as the molecules being matched were provided'''
    if rdmol_1.GetNumAtoms() != rdmol_2.GetNumAtoms():
        raise MolSizeMismatchError('Cannot generate bijection with Mols of two different sizes')
    
    atom_mapping = rdmol_1.GetSubstructMatch(rdmol_2) 
    if (not atom_mapping) or (len(atom_mapping) != rdmol_1.GetNumAtoms()): # TODO : verify whether the second condition necessary with the above size check
        raise SubstructMatchFailedError  # if no match or an incomplete match is found, raise Exception

    for rdatom_1_idx, rdatom_2 in zip(atom_mapping, rdmol_2.GetAtoms()):
        rdatom_2_idx = rdatom_2.GetIdx()
        yield (rdatom_1_idx, rdatom_2_idx)


"""A unified set of tools for setting up general organic polymer systems for molecular dynamics"""

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

# from ._version import __version__
from importlib.metadata import version
__version__ = version(__name__)

# used to check whether package is installed and importable
from .genutils import importutils

from .genutils.importutils.pkgiter import module_hierarchy
from .maths.combinatorics.tables import pascal


'''Utilities to streamline creation of 4x4 affine transformation matrices of 3D linear transformations in homogeneous coordinates'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import numpy as np
import numpy.typing as npt
from typing import Annotated, Literal


# USEFUL CUSTOM TYPEHINTS
Array4x4 = Annotated[npt.NDArray[np.float64], Literal[4, 4]]
AffineMatrix = Array4x4 # alias for convenience

# RIGID TRANSFORMATIONS
def xyzTrans(x : float=0.0, y : float=0.0, z : float=0.0) -> AffineMatrix:
    '''Returns rigid affine matrix which performs an isometric translation by "x", "y", and "z" units
    Returns the Identity matrix in absence of passed arguments'''
    return np.array([
        [1, 0, 0, x],
        [0, 1, 0, y],
        [0, 0, 1, z],
        [0, 0, 0, 1],
    ], dtype='float64')


def xRot(angle_rad : float=0.0) -> AffineMatrix:
    '''Returns rigid affine matrix which rotates about the positive x-axis by "angle_rad" radians 
    Returns the Identity matrix in absence of passed arguments'''
    s = np.sin(angle_rad)
    c = np.cos(angle_rad)
    
    return np.array([
        [1, 0,  0, 0],
        [0, c, -s, 0],
        [0, s,  c, 0],
        [0, 0,  0, 1],
    ], dtype='float64')

def yRot(angle_rad : float=0.0) -> AffineMatrix:
    '''Returns rigid affine matrix which rotates about the positive y-axis by "angle_rad" radians 
    Returns the Identity matrix in absence of passed arguments'''
    s = np.sin(angle_rad)
    c = np.cos(angle_rad)
    
    return np.array([
        [c, 0, -s, 0],
        [0, 1,  0, 0],
        [s, 0,  c, 0],
        [0, 0,  0, 1],
    ], dtype='float64')

def zRot(angle_rad : float=0.0) -> AffineMatrix:
    '''Returns rigid affine matrix which rotates about the positive z-axis by "angle_rad" radians 
    Returns the Identity matrix in absence of passed arguments'''
    s = np.sin(angle_rad)
    c = np.cos(angle_rad)
    
    return np.array([
        [c, -s, 0, 0],
        [s,  c, 0, 0],
        [0,  0, 1, 0],
        [0,  0, 0, 1],
    ], dtype='float64')

def randRot(about_x : bool=True, about_y : bool=True, about_z : bool=True) -> AffineMatrix:
    '''Returns an affine matrix for a rotation by some random angle(s) about the x, y, and z axes (or any subset of those axes)'''
    rot_dir = { # concise way to encapsulate what rotations can be performed and whther to perform them
        xRot : about_x,
        yRot : about_y,
        zRot : about_z,
    }

    matrix = np.eye(4, dtype=float)
    for (rot_fn, should_rotate) in rot_dir.items():
        if should_rotate:
            matrix = rot_fn(2 * np.pi * np.random.rand()) @ matrix # generate random angle and multiply rotation into overall transform matrix (in order)

    return matrix


# SCALING AND SHEAR
def xyzScale(sx : float=1.0, sy : float=1.0, sz : float=1.0) -> AffineMatrix:
    '''Returns rigid affine matrix which scales basis vectors by factors of "sx", "sy", and "sz" units
    Returns the Identity matrix in absence of passed arguments'''
    return np.array([
        [sx,  0,  0, 0],
        [ 0, sy,  0, 0],
        [ 0,  0, sz, 0],
        [ 0,  0,  0, 1],
    ], dtype='float64')

# TODO : add shear matrix functions


'''Abstract base and concrete implementations of algorithms which solve the DISCERNMENT Problem'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Generator, Generic, ParamSpec, Sequence, TypeVar
from abc import ABC, abstractmethod

from copy import deepcopy # TODO : deprecate the need for this in Cartesian
from itertools import product as cartesian_product

P = ParamSpec('P')
from .inventory import SymbolInventory, T, L


# HELPER FUNCTIONS
def is_unique(seq : Sequence) -> bool:
    '''Whether or not a Sequence contains repeating items'''
    return len(set(seq)) == len(seq)

# ABSTRACT BASE WHICH DEFINES STRATEGY INTERFACE
class DISCERNMENTStrategy(ABC, Generic[T, L]):
    '''
    DISCERNMENT = the Determination of Index Sequences from Complete Enumeration of Ransom Notes (Multiset Extension with Nonlexical Types)
    Provides an implementation of the general ransom note enumeration problem
    '''
    @abstractmethod
    def enumerate_choice_labels(self,
        word : Sequence[T],
        symbol_inventory : SymbolInventory[T, L],
        ignore_multiplicities : bool=False,
        unique_bins : bool=False,
        **kwargs : P.kwargs,
    ) -> Generator[tuple[L, ...], None, None]:
        '''
        Takes a word (N-element sequence of type T) and a symbol inventory (map from symbols and bin labels to counts),
        Exhaustively enumerates all N-tuples of indices corresponding to ordering of bins from which the word could be drawn
        
        Support modifications to base behavior:
        *** If ignore_multiplicities=True, will not respect the counts of elements in each bin when drawing
            For a given symbol, this allows a bin containing the symbol to appear more times than that symbol is present in that bin
        *** If unique_bins=True, will only allow each bin to be sampled from once, 
            EVEN if that bin contains symbols appearing later in the word
        '''
        pass

# CONCRETE IMPLEMENTATIONS
class DISCERNMENTStrategyCartesian(DISCERNMENTStrategy):
    '''
    Naive implementation of generalized ransom note enumeration strategy based on cartesian products
    
    Generates all possible index sequences ignoring multiplicity, then checks them one-by-one to see if they're valid

    Roughly 1-2 OOM slower (prefactor) than Recursive and Stack
    unless ignoring multiplicity in which case this is ~1 OOM faster
    '''
    def enumerate_choice_labels(self,
        word : Sequence[T],
        symbol_inventory : SymbolInventory[T, L],
        ignore_multiplicities : bool=False,
        unique_bins : bool=False,
    ) -> Generator[tuple[L, ...], None, None]:
        # generate every valid bin index tuple WITHOUT regard to repetition or uniqueness
        for indices in cartesian_product(*(symbol_inventory[symbol].keys() for symbol in word)): 
            if unique_bins and not is_unique(indices): 
                continue # skip non-unique bin choices if the option is set

            if not ignore_multiplicities:
                choice_inventory = deepcopy(symbol_inventory) # make a new copy for each path check
                for i, sym in zip(indices, word, strict=True):
                    if choice_inventory[sym][i] == 0:
                        overdrawn = True
                        break
                    choice_inventory[sym][i] -= 1
                else:
                    overdrawn = False 

                if overdrawn: # TOSELF: there are some optimizations to be made here (i.e. avoid copying inventory via backtrack)...
                    continue  # ... but am keeping this exact implementation for backward-compatibility reasons
            
            yield indices # only yield if all specified conditions are met

class DISCERNMENTStrategyRecursive(DISCERNMENTStrategy):
    '''
    Recursive implementation of generalized ransom note enumeration strategy via divide-and-conquer approach

    Breaks sequence down into first and all remaining symbols ("head" and "tail", respectively)
    Yields solution as all indices where the head occurs + recursive solutions to the tail sequence enumeration

    Easiest to analyze and reason about, but imposes cap on word length due to Python call stack size
    Performs intermediately in benchmark (faster than Cartesian, but slower than Stack)
    '''
    def enumerate_choice_labels(self,
        word : Sequence[T],
        symbol_inventory : SymbolInventory[T, L],
        ignore_multiplicities : bool=False,
        unique_bins : bool=False,
        _buffer : tuple[int]=None,
    ) -> Generator[tuple[L, ...], None, None]:
        if _buffer is None:
            _buffer = tuple()

        if not word: # base case for recursion
            yield _buffer # yields empty buffer if no sequence is present
            return

        symbol_cost = (0 if ignore_multiplicities else 1) # NOTE: definition here could be made more terse, but at the expense of readability
        
        symbol, *tail = word # separate head symbol from rest of sequence
        for bin_idx, num_occurences in symbol_inventory[symbol].items():
            # only proceed if letters are available from that bin, AND either uniqueness is not required, or it is required but has not yet been violated
            if (num_occurences > 0) and (not unique_bins or (bin_idx not in _buffer)):
                symbol_inventory[symbol][bin_idx] -= symbol_cost # mark current symbol and bin in symbol inventory and visit tracker, respectively

                yield from self.enumerate_choice_labels( # recursive traversal of remainder of sequence with current choice of bin for leading character
                    word=tail,
                    symbol_inventory=symbol_inventory,
                    ignore_multiplicities=ignore_multiplicities,
                    unique_bins=unique_bins,
                    _buffer=_buffer + (bin_idx,), # creates copy, rather than modifying the buffer for the current symbol (i.e. exactly what we want)
                )
                symbol_inventory[symbol][bin_idx] += symbol_cost # replace removed symbol for subsequent traversals to avoid polluting other states

class DISCERNMENTStrategyStack(DISCERNMENTStrategy):
    '''
    Stack-based implementation of generalized ransom note enumeration strategy

    Pushes all valid symbol-index pairs onto a stack for the current word positions,
    then advances and does the same until reaching the final symbol, backtracking and restoring symbols afterwards

    Fastest of all strategies across benchmark (except Cartesian **specifically** when ignore_multiplicities=True)
    '''
    def enumerate_choice_labels(self,
        word : Sequence[T],
        symbol_inventory : SymbolInventory[T, L],
        ignore_multiplicities : bool=False,
        unique_bins : bool=False,
    ) -> Generator[tuple[L, ...], None, None]:
        symbol_cost = (0 if ignore_multiplicities else 1) # NOTE: definition here could be made more terse, but at the expense of readability

        buffer = []
        bin_stack : list[tuple[T, int, bool]] = [(None, None, False)] # initialize with null symbol and bin, with no intent to replace (i.e. with intent to take)

        while bin_stack:
            symbol, bin_id, replace = bin_stack.pop()
            if replace:
                _last_bin_id = buffer.pop() # NOTE: this happens to be the same as bin_id (by construction) but will keep separate for clarity
                symbol_inventory[symbol][bin_id] += symbol_cost
            else:   
                if bin_id is not None: # exclude initial stack to give correct behavior in the case of an empty sequence (no explicit base case required!)
                    symbol_inventory[symbol][bin_id] -= symbol_cost
                    buffer.append(bin_id)

                if len(buffer) == len(word): # report buffer if the whole sequence has been accounted for
                    yield tuple(buffer)
                else: 
                    next_symbol = word[len(buffer)] # NOTE: symbol update fetches need to be done carefully WRT buffer updates
                    for next_bin_id, count in reversed(symbol_inventory[next_symbol].items()): # NOTE : reversal is solely to ensure overall order is consistent with other methods (needed since a stack is a FILO queue)
                        if (count > 0) and (not unique_bins or (next_bin_id not in buffer)):
                            bin_stack.append( (next_symbol, next_bin_id, True) )  # push replace operation onto stack first so it happens last
                            bin_stack.append( (next_symbol, next_bin_id, False) ) # presence of single take and replace


'''For generating and manipulating sequences of symbols which correspond to monomer ordering in blocky and random copolymers'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import logging
LOGGER = logging.getLogger(__name__)

from typing import Iterable, Optional
from dataclasses import dataclass, field, asdict

from ...genutils.textual.substrings import shortest_repeating_substring, repeat_string_to_length
from ...genutils.fileutils.jsonio.jsonify import make_jsonifiable
from ..exceptions import EndGroupDominatedChain, InsufficientChainLength, EmptyBlockSequence, PartialBlockSequence


@make_jsonifiable
@dataclass
class LinearCopolymerSequencer:
    '''
    For encapsulating information about the sequence of repeat units in a periodic, linear copolymer
    Also covers, as trivial special cases, homopolymers and alternating copolymers
    
    Parameters
    ----------
    sequence_kernel : str
        A sequence indicating a periodic ordering of monomers in a linear polymer block (e.g. "A", "ABAC", etc)
        Each unique symbol in the sequence corresponds to a distinct monomer in the block
    n_repeat_units : int
        The desired total number of monomers (including terminal monomers) in a polymer chain
    n_monomers_terminal : int
        The number of terminal monomers ("end groups") which are to be included in the chain
        in addition to the middle monomers described by "sequence"
        
    Raises
    ------
    EmpyBlockSequence
        The sequence provided is empty (can't be used to define nonzero-length chain)
    End GroupDominatedChain
        The number of terminal monomers exceed the number of total monomers
    '''
    sequence_kernel : str
    n_repeat_units : int
    n_repeat_units_terminal : int = 0
    
    # Attribute checks and modifications
    def __post_init__(self) -> None:
        if not self.sequence_kernel:
            raise EmptyBlockSequence('Must provide non-empty sequence kernel to yield a valid (co)polymer sequence')
    
        if self.n_repeat_units_middle < 0:
            raise EndGroupDominatedChain(
                f'Number of terminal monomers exceeds requested chain length; ({self.n_repeat_units}-mer ' \
                f'chain can\'t possibly contain {self.n_repeat_units_terminal} terminal monomers)'
            )
            
    def copy(self) -> 'LinearCopolymerSequencer':
        '''Returns another equivalent instance of the current sequence info more efficiently than a complete deepcopy'''
        return self.__class__(**asdict(self))
            
    def reduce(self) -> None:
        '''
        Determines if there is a shorter repeating subsequence making up the current sequence kernel
        If there is, adjusts the sequence kernel to that minimal sequence; does nothing otherwise
        
        Reduction is idempotent, and guarantees that the smallest possible kernel is used when sequencing
        '''
        minimal_subsequence = shortest_repeating_substring(self.sequence_kernel)
        kernel_period = self.block_size // len(minimal_subsequence) # account for any periodic shortening WITHIN the kernel
        
        if kernel_period == 1:
            LOGGER.info(f'Sequence kernel "{self.sequence_kernel}" is already fully reduced; no changes made')
            return
        else:
            LOGGER.info(
                f'Sequence kernel "{self.sequence_kernel}" can be further decomposed as {kernel_period}*"{minimal_subsequence}"; ' \
                f'Setting kernel to minimal subsequence "{minimal_subsequence}"'
            )
            self.sequence_kernel = minimal_subsequence
    
    def reduced(self) -> 'LinearCopolymerSequencer':
        '''Return a sequence-reduced version of the current sequence info'''
        clone = self.copy()
        clone.reduce()
        
        return clone
        
    # Properties derived from sequence kernel and target chain lengths
   
    @property
    def n_repeat_units_middle(self) -> int:
        '''Number of middle (i.e. non-terminal) repeat units'''
        return self.n_repeat_units - self.n_repeat_units_terminal

    # Whole sequence periods
    @property
    def block_size(self) -> int:
        '''Number of repeat units units in one whole iteration of the kernel block'''
        return len(self.sequence_kernel)
    period = block_size
    
    @property
    def n_full_periods(self) -> int:
        '''
        Largest number of complete repetitions of the sequence kernel which, when taken
        together, contain no more repeats units than the specified number of middle units
        '''
        return self.n_repeat_units_middle // self.block_size
    
    # Partial sequence residues
    @property
    def n_residual_repeat_units(self) -> int:
        '''
        Difference between number of middle repeat units and units which
        would occur in maximal full periods of the kernel
        
        By construction, is no greater than the block size and is
        identically zero exactly when a whole number of kernel repeats
        '''
        return self.n_repeat_units_middle % self.block_size
    n_residual_symbols = n_res = n_residual_repeat_units
    
    @property
    def has_residual(self) -> bool:
        '''
        Whether or not the target number of middle repeat units
        can be attained by a whole number of kernel repeats
        '''
        return bool(self.n_residual_repeat_units)
    
    @property
    def sequence_residual(self) -> str:
        '''Partial repeat of the kernel sequence needed to attain the speficied number of middle units'''
        return self.sequence_kernel[:self.n_residual_repeat_units]
    residual = sequence_residual
    
    ## PROCRUSTEAN sequence alignment
    def procrustean_alignment(self, allow_partial_sequences : bool=False) -> tuple[str, int]:
        '''
        PROCRUSTEAN: Periodic Recurrence Of Cyclic Repeat Unit Sequences, Truncated to an Exact and Arbitrary Number
        Stretches or truncates the sequence kernel to achieve a target sequence length, cycling through the kernel's period as many times as needed
        
        Algorithm produces a sequence string "P" and number of repeats "r" which, taken together, satisfy the following:
        - The number of units in r repeats of P plus the number of terminal monomers is precisely equal to the target number of monomers
        - The units in P cycle through the units in S, in the order they appear in S
        - The number of times S is cycled through in P is always a rational multiple of the length of S
        If no satisfiable sequence-count pair can be found, raises an appropriate informative exception
        '''
        if not self.has_residual: # the case where the target length happens to consist of a whole-number of repeats of the kernel
            if self.n_full_periods < 1: # NOTE: if it were up to me, this would be < 0 to allow dimers, but mBuild has forced my hand
                raise InsufficientChainLength(
                    f'{self.n_repeat_units}-monomer chain cannot accomodate both {self.n_repeat_units_terminal} end groups AND at least 1 middle monomer sequence'
                )
            sequence_procrustean = self.sequence_kernel
            n_seq_repeats = self.n_full_periods
        else:
            if not allow_partial_sequences:
                raise PartialBlockSequence(
                    f'Partial polymer block sequence required to meet target number of monomers ("{self.residual}" prefix of sequence "{self.sequence_kernel}");\n' \
                    'If this is acceptable, set "allow_partial_sequences=True" and try calling build routine again'
                )    
            sequence_procrustean = repeat_string_to_length(self.sequence_kernel, target_length=self.n_repeat_units_middle, joiner='')
            n_seq_repeats = 1 # just repeat the entire mixed-fraction length sequence (no full sequence repeats to exploit)
            
        return sequence_procrustean, n_seq_repeats
    
    def describe_order(self, end_group_names : Optional[Iterable[str]]=None, default_end_group_name : str='END-GROUP') -> str:
        '''Descriptive string presenting a condensed view of the order of repeat units in the final sequence'''
        # Assign names for end groups
        if end_group_names is None:
            end_group_names = [f'[{default_end_group_name}]']*self.n_repeat_units_terminal 
        else:
            end_group_names = [f'[{end_group_name}]' for end_group_name in end_group_names] # unpack into list and enforce correct number of names
        if (num_names_provided := len(end_group_names)) != self.n_repeat_units_terminal: # DEV: consider supporting filling in missing names with default in future
            raise IndexError(f'Defined sequence info with {self.n_repeat_units_terminal} end groups, but only provided names for {num_names_provided}')
        
        # Insert middle omnomer parts as necessary
        sequence_middle = []
        if self.n_full_periods != 0: ## Whole sequence strings
            sequence_middle.append(f'{self.n_full_periods}*[{self.sequence_kernel}]')
        if self.has_residual: ## Partial sequence strings
            sequence_middle.append(f'[{self.residual}]')
            
        # Abut with correct amount of end group indicators
        sequence_parts = end_group_names[:] 
        sequence_parts[1:-1] = sequence_middle
        
        return ' + '.join(sequence_parts)
    
    def describe_tally(self) -> str:
        '''Descriptive string indicating how all parts of the overall sequence contribute to the target number of repeat units'''
        desc_seq_counts_parts = []
        if self.n_full_periods != 0: ## Whole sequence strings
            desc_seq_counts_parts.append(f'{self.n_full_periods} whole {self.block_size}-sequence repeat(s)')
        if self.has_residual: ## Partial sequence strings
            desc_seq_counts_parts.append(f'a partial {self.n_residual_repeat_units}/{self.block_size} sequence repeat')
            
        return ' and '.join(desc_seq_counts_parts)
        


'''For handling interconversion between the OpenMM and OpenFF (Pint) unit engines'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Callable, TypeVar

ReturnType = TypeVar('ReturnType') 

from openmm.unit import Quantity as OpenMMQuantity
from pint import Quantity as PintQuantity # this is also the base class for all OpenFF-style units

from openff.units.openmm import (
    from_openmm as openmm_to_openff,
    to_openmm as openff_to_openmm,
)


def allow_openmm_units(funct : Callable[..., ReturnType]) -> Callable[..., ReturnType]:
    '''Allow a Callable which expects any of its args to be OpenFF Quantities to also accept equivalent OpenMM Quantites'''
    def wrapper(*args, **kwargs) -> ReturnType:
        new_args = [
            openmm_to_openff(arg) if isinstance(arg, OpenMMQuantity) else arg
                for arg in args
        ]

        new_kwargs = {
            key : openmm_to_openff(kwarg) if isinstance(kwarg, OpenMMQuantity) else kwarg
                for key, kwarg in kwargs.items()
        }

        return funct(*new_args, **new_kwargs)
    return wrapper

def allow_openff_units(funct : Callable[..., ReturnType]) -> Callable[..., ReturnType]:
    '''Allow a Callable which expects any of its args to be OpenMM Quantities to also accept equivalent OpenFF Quantites'''
    def wrapper(*args, **kwargs) -> ReturnType:
        new_args = [
            openff_to_openmm(arg) if isinstance(arg, PintQuantity) else arg
                for arg in args
        ]

        new_kwargs = {
            key : openff_to_openmm(kwarg) if isinstance(kwarg, PintQuantity) else kwarg
                for key, kwarg in kwargs.items()
        }

        return funct(*new_args, **new_kwargs)
    return wrapper

# TODO : add decorators which allows OUTPUTS to have flexible units


'''Utilities for calculating fundamental combinatorial numbers'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Iterable
from operator import mul
from math import factorial # not worth re-implementing here, the C-implementation is plenty fast
from functools import lru_cache, reduce


# Sequential utility functions
def alt_sign(n : int) -> int:
    '''Returns the value of -1^n (useful for alternating sums)'''
    # return 2*(n % 2) - 1 # arithmetic turns out to be slightly slower than explicit check
    return -1 if (n % 2) else 1 


# Binomial and multinomial coefficients
@lru_cache
def binomial_coeff(n : complex, k : int) -> float: # appears to be quicker than scipy.special.binom
    '''Calculates a generalized binomial coefficient
    Counts the number of distinct ways to make and unordered selection of k elemnts from a set of n items'''
    if (k < 0):
        return 0.0
    
    if (k > n//2) and isinstance(n, int):
        return binomial_coeff(n, n - k)

    bincoeff = 1.0
    for i in range(k):
        bincoeff *= (n - i) / (k - i)

    return bincoeff
choose = binomial_coeff # alias for convenience

def multiset_coeff(n : int, k : int) -> float:
    '''Calculates a multiset coefficient
    Counts the number of ways to form a multiset of cardinality n from a support of size k'''
    return binomial_coeff(n + k -1, k)

def multinomial_coeff(multips : Iterable[int]) -> float:
    '''Calculate the multinomial coefficient of a set of multiplicities
    Counts the number of distinct permutations of a multiset with given multiplicites'''
    num = factorial(sum(multips))
    denom = reduce(mul, (factorial(i) for i in multips)) 

    return num / denom

def multinomial_coeff_native(multips : Iterable[int]) -> float:
    '''Calculate the multinomial coefficient of a set of multiplicities
    Counts the number of distinct permutations of a multiset with given multiplicites

    Native Python implementation of multinomial_coeff() which is slightly slower but requires no imports'''
    multicoeff, total = 1, 0 # initialize with multiplicate and additive identities, respectively
    for elem in multips:     # use expansion as product of binomials
        total += elem
        multicoeff *= binomial_coeff(total, elem)

    return multicoeff

def catalan(n : int) -> float:
    '''Calculates the n-th Catalan number
    Counts dozens of different things, including the number of monotonic 2D random walks
    and the number of valid (i.e. fully closed) groupings of n pairs of parentheses'''
    return binomial_coeff(2*n, n) / (n + 1)


# Counting partitions and cycles
def pentagonal(n : int) -> float:
    '''Calculates the n-th pentagonal number
    Counts the number of points in n "concentric" pentagon figures which share a corner vertex

    Related to the generating function for numbers of partitions'''
    # return binomial_coeff(n, 1) + 3*binomial_coeff(n, 2)
    return (3*n**2 - n) / 2

@lru_cache
def stirling_second(n : int, k : int) -> float:
    '''Calculates the Stirling number of the second kind S(n, k)
    Counts the number of ways to partition n objects into k non-empty pairwise-disjoint subsets

    Recurrence used as listed in NIST DLMF (https://dlmf.nist.gov/26.8)'''
    return sum(
        alt_sign(k - i) * binomial_coeff(k, i) * i**n
            for i in range(k + 1)
    ) / factorial(k)

@lru_cache
def stirling_first(n : int, k : int) -> float: # note : needs to be defined second, as the Stirling numbers of the second kinds are part of the definition here
    '''Calculates the Stirling number of the first kind c(n, k)
    Counts the number of ways to to decompose a permutation of n objects into k cycles

    Recurrence used as listed in Abramowitz ans Stegun (https://www.convertit.com/Go/ConvertIt/Reference/AMS55.ASP?Res=150&Page=825)'''
    if (n == 0) and (k == 0):
        return 1.0 # for some reason, the recurrence given does not give the correct boundary value

    return sum(
        alt_sign(j - k) * binomial_coeff(j - 1, k - 1) * binomial_coeff(2*n - k, j) * stirling_second(j - k, j - n)
            for j in range(n, 2*n - k + 1)
    )
    
@lru_cache
def bell(n : int) -> float:
    '''Calculates the n-th Bell number
    Counts the number of ways to partition n objects into any number of non-empty pairwise-disjoint subsets'''
    return sum(
        stirling_second(n, k)
            for k in range(n + 1)
    )

@lru_cache
def bernoulli(n : int) -> int: # NOTE: avoided implementing in terms of Stirling numbers of second kind due to catastophic cancellation of factorials
    '''Calculates the n-th Bernoulli number
    Related to the sums of powers of integers (via Faulhalber's formula)'''
    if n == 0:
        return 1
    return -sum( # NOTE: implemented recursively, as both explicit and Stirling number (2nd) implementations fail even for n < 20
        binomial_coeff(n + 1, k) * bernoulli(k)
            for k in range(n)
    ) / (n + 1)

# Graph enumeration
@lru_cache
def count_labelled_graphs(N : int) -> int:
    '''Gives number of all labelled graphs on N nodes'''
    return 2**binomial_coeff(N, 2)

@lru_cache
def count_connected_labelled_graphs(n : int) -> float:
    '''Gives array of number labelled graphs with one connected component on N nodes'''
    if n == 0:
        return 0.0 # TOSELF : should this be 1 instead? (Really more of a philosophical question, should the null graph be counted as having 1 component?)
    return count_labelled_graphs(n) - sum(
        binomial_coeff(n - 1, k - 1)*count_labelled_graphs(n - k)*count_connected_labelled_graphs(k)
            for k in range(1, n)
    )


'''Type aliases for Python callable input parameters'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import TypeVar, ParamSpec, _UnionGenericAlias

T = TypeVar('T') # universal generic type

C = TypeVar('C') # generic type for a class
O = TypeVar('O') # generic type for an object passed to a function
F = TypeVar('F') # generic type for a function
U = TypeVar('U', bound=_UnionGenericAlias) # generic class from representing Literal Unions (since arg-free union is not supported)

P = ParamSpec('P') # for representing (preserved) input parameters
R = TypeVar('R')   # for representing generic return values

# TOSELF: replace with typing.ParamSpecArgs/.ParamSpecKWArgs?
Args   = TypeVar('Args'  ) # generic type for arguments to a function
KWArgs = TypeVar('KWArgs') # generic type for keyword arguments to a function



'''Core tools for manipulating integer lattices in D-dimensions'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Iterable
from numbers import Number
from ...genutils.typetools.numpytypes import Shape, D, N

import numpy as np
from itertools import product as cartesian_product
from .coordinates import Coordinates, BoundingBox


def generate_int_lattice(*dims : Iterable[int]) -> np.ndarray[Shape[N, D], int]:
    '''Generate all D-D coordinates of points on a integer lattice with the sizes of all D dimensions given'''
    return np.fromiter(
        iter=cartesian_product(*[
            range(d)
                for d in dims
        ]),
        dtype=np.dtype((int, len(dims)))
    )

def nearest_int_coord_along_normal(point : np.ndarray[Shape[D], Number], normal : np.ndarray[Shape[D], Number]) -> np.ndarray[Shape[D], int]:
    '''
    Takes an D-dimensional control point and an D-dimensional normal vector
    Returns the integer-valued point nearest to the control point which lies in
    the normal direction relative to the control point
    '''
    # Check that inputs have vector-like shapes and compatible dimensions
    assert(point.ndim == 1)
    assert(normal.ndim == 1)
    assert(point.size == normal.size)

    min_int_bound_point = np.floor(point)
    max_int_bound_point = np.ceil( point)
    if np.isclose(min_int_bound_point, max_int_bound_point).all(): # edge case: if already on an integer-valued point, the fllor and ceiling will be identical
        int_point = point 
    else:
        int_bbox = BoundingBox(np.vstack([min_int_bound_point, max_int_bound_point]))  # generate bounding box from extremal positions
        dots = np.inner((int_bbox.vertices - point), normal) # take dot product between the normal and the direction vectors from the control point to the integer bounding points
        i = dots.argmax() # position of integer point in most similar direction to normal
        furthest_point, similarity = int_bbox.vertices[i], dots[i]

        if similarity <= 0:
            raise ValueError(f'Could not locate valid integral point in normal direction (closest found was {furthest_point} with dot product {similarity})')
        else:
            int_point = furthest_point 

    return int_point.astype(int)

# TODO : implement enumeration of integral points within an D-simplex

class CubicIntegerLattice(Coordinates[int]):
    '''For representing an n-dimensional integer lattice, consisting of all n-tuples of integers with values constrained by side lengths in each dimension'''
    def __init__(self, counts_along_dims : np.ndarray[Shape[D], int]) -> None: # TODO: implement more flexible input support (i.e. star-unpacking, listlikes, etc.)
        assert(counts_along_dims.ndim == 1)
        super().__init__(generate_int_lattice(*counts_along_dims))
        self.counts_along_dims = counts_along_dims

    def counts_along_dims_as_str(self, multip_char : str='x') -> str:
        '''Stringify the lattice sidelengths'''
        return multip_char.join(str(i) for i in self.counts_along_dims)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.n_dims}-dimensional, {self.counts_along_dims_as_str()})'

    # LATTICE DIMENSIONS
    @property
    def capacity(self) -> int: # referred to as "N" in typehints
        '''The maximum number of points that the lattice could contains'''
        return np.prod(self.counts_along_dims)

    @property
    def lex_ordered_weights(self) -> np.ndarray[Shape[D], int]:
        '''Vector of the number of points corresponding
        Can be viewed as a linear transformation between indices and point coordinates when coords are placed in lexicographic order'''
        return np.concatenate(([1], np.cumprod(self.counts_along_dims)[:-1]))

    # SUBLATTICE DECOMPOSITION
    @property
    def odd_even_idxs(self) -> tuple[np.ndarray[Shape[N], int], np.ndarray[Shape[N], int]]: # TOSELF: each subarray actually has length N/2 (+1 if capacity is odd), not sure how to typehint that though
        '''Return two vectors of indices, corresponding to points in the "odd" and "even" non-neighboring sublattices, respectively'''
        parity_vector = np.mod(self.points.sum(axis=1), 2) # remainder of sum of coordinates of each point; corresponds to the condition that a single step along any dimension should invert parity
        is_odd = parity_vector.astype(bool) # typecast as booleans to permit indexing (and make intent a bit clearer)

        return np.flatnonzero(is_odd), np.flatnonzero(~is_odd) # need to faltten to avoid inconvenient tuple wrapping

    @property
    def odd_idxs(self) -> np.ndarray[Shape[N], int]:
        '''Indices of the point in the in "odd" sublattice'''
        return self.odd_even_idxs[0]

    @property
    def even_idxs(self) -> np.ndarray[Shape[N], int]:
        '''Indices of the point in the in "even" sublattice'''
        return self.odd_even_idxs[1]

    @property
    def odd_sublattice(self) -> np.ndarray[Shape[N, D], int]:
        '''Returns points within the odd sublattice of the lattice points'''
        return self.points[self.odd_idxs]
    odd_points = odd_sublattice # alias for convenience

    @property
    def even_sublattice(self) -> np.ndarray[Shape[N, D], int]:
        '''Returns points within the even sublattice of the lattice points'''
        return self.points[self.even_idxs]
    even_points = even_sublattice # alias for convenience


'''Conversion tools for representing positive integers in fixed and mixed radix positional bases'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Any, Generator, Sequence, Union
from math import inf


class PositionalNumberingSystem:
    '''For representing positive integers in standard and mixed-radix positional numbering systems'''
    def __init__(self, radices : Sequence[int]) -> None:
        '''Radices should be passed in ascending order of significance (i.e. least-significant bases first)'''
        self._radices = radices # add check for integers
     
    # descriptor for place values access (read-only)
    @property
    def radices(self) -> Sequence[int]:
        return self._radices
    bases = radices
    # NOTE: specifically opted not to include an iter() conversion property, as this would reset a generator on each call inside the loop

    @property
    def places(self) -> Generator[int, None, None]:
        '''Generates the values of each position in a general mixed-radix positional notation'''
        p = 1 # TOSELF: should also yield 1?
        for radix in self.radices:
            yield p
            p *= radix

    # implementations of general base conversions
    def int_to_digits_iter(self, n : int) -> Generator[int, None, None]:
        '''Convert a non-negative integer to its digit representation under the specified radices
        Yields digits in ascending order of significance (i.e. starting from least significant digit)'''
        if n < 0:
            raise ValueError('Cannot generate digits of a negative integer')

        radices = iter(self.radices) # convert to iterable to generator-like parsing of sequences
        while n:
            try:
                radix = next(radices)
            except StopIteration: # if radices have run out, yield residual as "infinitieth" digit
                yield n
                break

            n, digit = divmod(n, radix)
            if digit < 0: # account for downwards rounding when encountering negative radices
                n += 1
                digit -= radix

            yield digit

    def int_to_digits(self, n : int, as_str : bool=False) -> Union[str, list[int]]:
        '''Returns digits of number in descending order of significance (i.e. most significant digit first)'''
        digits = [digit for digit in self.int_to_digits_iter(n)]
        digits.reverse()
        
        if as_str:
            return ''.join(str(digit) for digit in digits)
        return digits

    def digits_to_int(self, digits : Sequence[int]) -> int:
        '''Convert a sequence of digits to an integer in the specified numbering system
        Digits should be passed in descending order (i.e. most-significant digit first)'''
        n : int = 0

        radices = iter(self.radices)
        for i, digit, place in enumerate(zip(reversed(digits), self.places), start=0):
            digit = int(digit) # allows for string passing

            try:
                radix = next(radices)
            except StopIteration:# radix is deliberately one position ahead in sequence, to allow "peek" ahead due to 
                radix = inf
            
            if not (0 <= digit <= radix):
                raise ValueError(f'Digit {digit} in position {i} exceeds max allowed digit value of {radix} for that place')
            n += digit*place
        
        return n 

    # dunders for concise calling
    def __call__(self, *args, **kwargs) -> list[int]:
        '''Syntactic sugar for digit-to-int conversion (expects a Sequence)'''
        if (len(args) == 1) and isinstance(args[0], Sequence):
            args = args[0]
        return self.digits_to_int(args, **kwargs)
            
    def __getitem__(self, value : Any) -> int:
        '''Syntactic sugar for int-to-digit conversion'''
        if isinstance(value, int):
            return self.int_to_digits(value, as_str=True)
        else:
            raise TypeError



'''Generic operations for indexing, generating, and iterating over sequences'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Generator, Sequence, TypeVar, Union
T = TypeVar('T') # generic type for sequence element
S = TypeVar('S') # generic type for a distinct sequence element
from itertools import count


def is_unique(seq : Sequence) -> bool:
    '''Whether or not a Sequence contains repeating items'''
    return len(set(seq)) == len(seq)

def int_complement(integers : Sequence[int], bounded : bool=False) -> Generator[int, None, None]:
    '''Generate ordered non-negative integers which don't appear in a sequence of integers'''
    _max = max(integers) # cache maximum (precludes use of generator-like sequence)
    for i in range(_max):
        if i not in integers:
            yield i

    if not bounded: # keep counting past max if unbounded
        yield from count(start=_max + 1, step=1)

def pad_sequence(target_list : Sequence[T], to_length : int, pad_value : S=0, from_left : bool=False) -> list[Union[T, S]]:
    '''Pad a given list with a particular value'''
    padding_list = [pad_value] * (to_length - len(target_list)) # will be empty if the target length is shorter than the provided list (i.e. no padding)
    if from_left:
        return padding_list + target_list
    return target_list + padding_list

def cycle_items(seq : Sequence[T], places : int=1) -> list[T]:
    '''
    Cyclically shift all items in a sequence over by the target number of indices
    By default shifts right, but can also move to left with negative "places" argument
    
    Examples:
        cycle_items([1,2,3,4],  2) -> [3,4,1,2]
        cycle_items([1,2,3,4], -1) -> [4,1,2,3]
    '''
    n_items = len(seq) # this length call is what requires the input to be a Sequence and not just an Iterable
    return [
        seq[i % n_items]
            for i in range(places, places + n_items)
    ]


'''For assigning, transferring, and removing properties of RDKit objects'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Any, Optional, TypeVar, TypeAlias, Union
from copy import deepcopy

from rdkit import Chem
from rdkit.Chem.rdchem import Atom, Bond, Mol, RWMol

from .labeling.bijection import bijective_atom_id_iter
from ..genutils.decorators.functional import optional_in_place
from ..genutils.typetools.categorical import _union_member_factory
from ..genutils.attrs import compile_argfree_getable_attrs


# RDKit-specific generics and type aliases
T = TypeVar('T')   # generic type for AtomProp attributes
RD = TypeVar('RD') # generic type to represent an RDKit object

RDObj : TypeAlias = Union[Atom, Bond, Mol, RWMol]
isrdobj = _union_member_factory(RDObj, 'RDObj')


# REFERENCE FOR "MAGIC" PROP KEYS AND DESCRIPTIONS IN RDKit (from https://www.rdkit.org/docs/RDKit_Book.html#magic-property-values) 
RDMOL_MAGIC_PROPS = {
    'MolFileComments'        : 'Read from/written to the comment line of CTABs.',
    'MolFileInfo'            : 'Read from/written to the info line of CTABs.',
    '_MolFileChiralFlag'     : 'Read from/written to the chiral flag of CTABs.',
    '_Name'                  : 'Read from/written to the name line of CTABs.',
    '_smilesAtomOutputOrder' : 'The order in which atoms were written to SMILES',
    '_smilesBondOutputOrder' : 'The order in which bonds were written to SMILES',
}

RDATOM_MAGIC_PROPS = {
    '_CIPCode'               : 'the CIP code (R or S) of the atom',
    '_CIPRank'               : 'the integer CIP rank of the atom',
    '_ChiralityPossible'     : 'set if an atom is a possible chiral center',
    '_MolFileRLabel'         : 'integer R group label for an atom, read from/written to CTABs.',
    '_ReactionDegreeChanged' : 'set on an atom in a product template of a reaction if its degree changes in the reaction',
    '_protected'             : 'atoms with this property set will not be considered as matching reactant queries in reactions',
    'dummyLabel'             : '(on dummy atoms) read from/written to CTABs as the atom symbol',
    'molAtomMapNumber'       : 'the atom map number for an atom, read from/written to SMILES and CTABs',
    'molfileAlias'           : 'the mol file alias for an atom (follows A tags), read from/written to CTABs',
    'molFileValue'           : 'the mol file value for an atom (follows V tags), read from/written to CTABs',
    'molFileInversionFlag'   : 'used to flag whether stereochemistry at an atom changes in a reaction, read from/written to CTABs, determined automatically from SMILES',
    'molRxnComponent'        : 'which component of a reaction an atom belongs to, read from/written to CTABs',
    'molRxnRole'             : 'which role an atom plays in a reaction (1=Reactant, 2=Product, 3=Agent), read from/written to CTABs',
    'smilesSymbol'           : 'determines the symbol that will be written to a SMILES for the atom',
}


# REFERENCE TABLES FOR ENFORCING C++ TYPING THAT RDKit ENFORCES
RDPROP_GETTERS = {
    str   : 'GetProp',
    bool  : 'GetBoolProp',
    int   : 'GetIntProp',
    float : 'GetDoubleProp'
}
RDPROP_SETTERS = {
    str   : 'SetProp',
    bool  : 'SetBoolProp',
    int   : 'SetIntProp',
    float : 'SetDoubleProp'
}

# PROPERTY INSPECTION FUNCTIONS
def detailed_rdobj_info(rdobj : RDObj) -> dict[str, Any]:
    '''Extract all get-able info about a particular RDKit atom. Does NOT include any non-default Prop values (e.g. atomMapNumber)'''
    return compile_argfree_getable_attrs(rdobj, getter_re='Get', repl_str='')

def atom_ids_with_prop(rdmol : Mol, prop_name : str) -> list[int]:
    '''Returns list of atom IDs of atom which have a particular property assigned'''
    return [
        atom.GetIdx()
            for atom in rdmol.GetAtoms()
                if atom.HasProp(prop_name)
    ]

@optional_in_place
def annotate_atom_prop(rdmol : Mol, prop : str, prop_type : T=str, annotate_precision : Optional[int]=None) -> None:
    '''Labels the desired Prop for all atoms in a Mol which have it'''
    getter_type = RDPROP_GETTERS[prop_type]
    for atom in rdmol.GetAtoms():
        prop_val = getattr(atom, getter_type)(prop) # invoke type-appropriate getter on atom, with the name of the desired property
        
        if hasattr(prop_val, '__round__') and annotate_precision is not None: # only round on roundable objects, and only when
            prop_val = round(prop_val, annotate_precision)
        atom.SetProp('atomNote', str(prop_val)) # need to convert to string, as double is susceptible to float round display errors (shows all decimal places regardless of rounding)

def aggregate_atom_prop(rdmol : Mol, prop : str, prop_type : T=str) -> dict[int, T]:
    '''Collects the values of a given Prop across all atoms in an RDKit molecule'''
    getter_type = RDPROP_GETTERS[prop_type]
    return {
        atom_idx : getattr(rdmol.GetAtomWithIdx(atom_idx), getter_type)(prop) # invoke type-appropriate getter on atom, with the name of the desired property
            for atom_idx in atom_ids_with_prop(rdmol, prop_name=prop)
    }


# PROPERTY TRANSFER FUNCTIONS
def copy_rd_props(from_rdobj : RD, to_rdobj : RD) -> None: # NOTE : no need to incorporate typing info, as RDKit objects can correctly interpret typed strings
    '''For copying properties between a pair of RDKit Atoms or Mols'''
    # NOTE : avoid use of GetPropsAsDict() to avoid errors from restrictive C++ typing
    assert isrdobj(from_rdobj) and isrdobj(to_rdobj) # verify that both objects passed are RDKit objects...
    assert type(from_rdobj) == type(to_rdobj)        # ...AND that both objects are the same type of RDKit object

    for prop in from_rdobj.GetPropNames():
        to_rdobj.SetProp(prop, from_rdobj.GetProp(prop))

@optional_in_place
def assign_props_from_dict(prop_dict : dict[str, Any], rdobj : RDObj, preserve_type : bool=True) -> None:
    '''Copies all attributes from a strng-keyed dictionary of properties as Props of an RDKit object'''
    for key, value in prop_dict.items():
        if (type(value) not in RDPROP_SETTERS) or (not preserve_type): # set as string if type is unspecified or if explicitly requested to
            rdobj.SetProp(key, str(value))
        else:
            setter = getattr(rdobj, RDPROP_SETTERS[type(value)]) # use the atom's setter for the appropriate type
            setter(key, value) # pass key and value to setter method


# PROPERTY REMOVAL FUNCTIONS
@optional_in_place
def clear_atom_props(rdmol : Mol) -> None:
    '''Wipe properties of all atoms in a molecule'''
    for atom in rdmol.GetAtoms():
        for prop_name in atom.GetPropNames():
            atom.ClearProp(prop_name)


# PROPERTY COMPARISON FUNCTIONS
def difference_rdmol(rdmol_1 : Mol, rdmol_2 : Mol, prop : str='PartialCharge', remove_map_nums : bool=True) -> Mol:
    '''
    Takes two RDKit Mols (presumed to have the same structure and atom map numbers) and the name of a property 
    whose partial charges are the differences betwwen the two Mols' charges (atomwise)
    
    Assumes that the property in question is numeric (i.e. can be interpreted as a float)
    '''
    diff_mol = deepcopy(rdmol_1) # duplicate first molecule as template
    all_deltas = []
    for rdatom_1_idx, rdatom_2_idx in bijective_atom_id_iter(rdmol_1, rdmol_2):
        rdatom_1 = rdmol_1.GetAtomWithIdx(rdatom_1_idx)
        rdatom_2 = rdmol_2.GetAtomWithIdx(rdatom_2_idx)
        diff_atom = diff_mol.GetAtomWithIdx(rdatom_1_idx) # same index, since it is a deep copy

        delta = rdatom_1.GetDoubleProp(prop) - rdatom_2.GetDoubleProp(prop)
        diff_atom.SetDoubleProp(f'Delta{prop}', delta)
        all_deltas.append(delta)

        diff_atom.ClearProp(prop) # reset property value from original atom copy to avoid confusion
        if remove_map_nums:
            diff_atom.ClearProp('molAtomMapNumber') # Remove atom map num for greater visual clarity when drawing

    diff_mol.ClearProp(prop) # reset property value from original mol copy to avoid confusion
    diff_mol.SetProp(f'Delta{prop}s', str(all_deltas)) # label stringified version of property list (can be de-stringified via ast.literal_eval)
    diff_mol.SetDoubleProp(f'Delta{prop}Min', min(all_deltas)) # label minimal property value for ease of reference
    diff_mol.SetDoubleProp(f'Delta{prop}Max', max(all_deltas)) # label maximal property value for ease of reference

    return diff_mol


'''Tools for iterating over and extracting information from Python package hierarchies'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import logging
LOGGER = logging.getLogger(__name__)

from types import ModuleType
from typing import Generator, Iterable, Optional, Union

from importlib import import_module
from pkgutil import iter_modules as _iter_modules

from anytree.node import Node
from anytree.render import AbstractStyle, ContStyle
from anytree.iterators import PreOrderIter

from .pkginspect import module_stem, is_package
from .dependencies import MissingPrerequisitePackage
from ..trees.treebase import NodeCorrespondence, compile_tree_factory
from ..trees.treeviz import treestr



# HIERARCHICAL MODULE TREE GENERATION
class ModuleToNodeCorrespondence(NodeCorrespondence, FROMTYPE=ModuleType):
    '''Concrete implementation of Python modules and packages as nodes in a tree'''
    def name(self, module : ModuleType) -> str:
        return module_stem(module) # TODO: find Python package-spec compliant way of extracting this easily
    
    def has_children(self, module : ModuleType) -> bool:
        return is_package(module)
    
    def children(self, module : ModuleType) -> Iterable[ModuleType]:
        for _loader, module_name, ispkg in _iter_modules(module.__path__, prefix=module.__name__+'.'):
            try:
                submodule = import_module(module_name)
                yield submodule
            except (ModuleNotFoundError, MissingPrerequisitePackage):
                continue

module_tree = compile_tree_factory(
    ModuleToNodeCorrespondence(),
    class_alias='package',
    obj_attr_name='module',
    exclude_mixin=lambda module : module_stem(module).startswith('_'),
)

# BACKWARDS-COMPATIBLE PORTS OF LEGACY IMPORTUTILS FUNCTIONS
def module_tree_direct(module : ModuleType, recursive : bool=True, blacklist : Optional[Iterable[str]]=None) -> Node:
    '''Produce a tree from the Python package hierarchy starting with a given module
    
    Parameters
    ----------
    module : ModuleType
        The "root" module to begin importing from
        Represented in the Node object returned by this function
    recursive : bool, default=True
        Whether or not to recursively import modules from subpackages and add them to the tree
    blacklist : list[str] (optional), default None
        List of module names to exclude from tree building
        If provided, will exclude any modules whose names occur in this list

    Returns
    -------
    modtree : Node
        The root node of the module tree, corresponding to the module object passed to "module"
    '''
    if blacklist is None:
        blacklist = []

    return module_tree(
        module,
        max_depth=None if recursive else 1,
        exclude=lambda module : module_stem(module) in blacklist,
    )

def iter_submodules(module : ModuleType, recursive : bool=True, blacklist : Optional[Iterable[str]]=None) -> Generator[ModuleType, None, None]:
    '''
    Generates all modules which can be imported from the given toplevel module
    
    Parameters
    ----------
    module : ModuleType
        The "root" module to begin importing from
        Represented in the Node object returned by this function
    recursive : bool, default=True
        Whether or not to recursively import modules from subpackages and add them to the tree
    blacklist : list[str] (optional), default None
        List of module names to exclude from tree building
        If provided, will exclude any modules whose names occur in this list

    Returns
    -------
    submodules : Generator[ModuleType]
        A generator which yields modules in traversal pre-order as they appear wihin the package hierarchy
    '''
    modtree = module_tree_direct(module, recursive=recursive, blacklist=blacklist)
    for module_node in PreOrderIter(modtree):
        yield module_node.module

def iter_submodule_info(module : ModuleType, recursive : bool=True, blacklist : Optional[Iterable[str]]=None) -> Generator[tuple[ModuleType, str, bool], None, None]:
    '''
    Generates information about all modules which can be imported from the given toplevel module
    Namely, yields the module object, module name, and whether or not the module is a package

    Parameters
    ----------
    module : ModuleType
        The "root" module to begin importing from
        Represented in the Node object returned by this function
    recursive : bool, default=True
        Whether or not to recursively import modules from subpackages and add them to the tree
    blacklist : list[str] (optional), default None
        List of module names to exclude from tree building
        If provided, will exclude any modules whose names occur in this list

    Returns
    -------
    submodule_info : Generator[ModuleType, str, bool]
        A generator which yields modules info in traversal pre-order as they appear wihin the package hierarchy
        yields 3-tuples containing ModuleType objects, module names, and whether the current module is also a subpackage
    '''
    modtree = module_tree_direct(module, recursive=recursive, blacklist=blacklist)
    for module_node in PreOrderIter(modtree):
        yield module_node.module, module_node.name, module_node.is_leaf

def register_submodules(module : ModuleType, recursive : bool=True, blacklist : Optional[Iterable[str]]=None) -> None:
    '''
    Registers all submodules of a given module into it's own namespace (i.e. autoimports submodules)
    
    Parameters
    ----------
    module : ModuleType
        The "root" module to begin importing from
        Represented in the Node object returned by this function
    recursive : bool, default=True
        Whether or not to recursively import modules from subpackages and add them to the tree
    blacklist : list[str] (optional), default None
        List of module names to exclude from tree building
        If provided, will exclude any modules whose names occur in this list

    Returns
    -------
    None
    '''
    for submodule in iter_submodules(module, recursive=recursive, blacklist=blacklist):
        setattr(module, submodule.__name__, submodule)

def module_hierarchy(module : ModuleType, recursive : bool=True, blacklist : Optional[Iterable[str]]=None, style : Union[str, AbstractStyle]=ContStyle()) -> str:
    '''
    Generates a printable string which summarizes a Python packages hierarchy. Reminiscent of GNU tree output

    Parameters
    ----------
    module : ModuleType
        The "root" module to begin importing from
        Represented in the Node object returned by this function
    recursive : bool, default=True
        Whether or not to recursively import modules from subpackages and add them to the tree
    blacklist : list[str] (optional), default None
        List of module names to exclude from tree building
        If provided, will exclude any modules whose names occur in this list
    style : str or AbstractStyle
        An element drawing style for the final tree structure printout

    Returns
    -------
    module_summary : str
        Printable string which displays the package structure
    '''
    modtree = module_tree_direct(module, recursive=recursive, blacklist=blacklist)
    return treestr(modtree, style=style)




'''Utilities for generating, storing, and applying partial charges to OpenFF Molecules'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import logging
LOGGER = logging.getLogger(__name__)

from typing import Optional

import numpy as np
from collections import defaultdict
from rdkit import Chem

from openff.toolkit.topology.molecule import Molecule
from openmm.unit import elementary_charge

from .rctypes import ChargesByResidue
from .redistribution import ChargeRedistributionStrategy, UniformDistributionStrategy
from .....genutils.decorators.functional import optional_in_place
from .....maths.statistics import Accumulator
from .....polymers.monomers.repr import MonomerGroup


# FUNCTIONS FOR RESIDUE CHARGE CALCULATION
def find_repr_residues(offmol : Molecule) -> dict[str, int]:
    '''Determine names and smallest residue numbers of all unique residues in charged molecule
    Used as representatives for generating labelled SMARTS strings '''
    rep_res_nums = defaultdict(set) # numbers of representative groups for each unique residue, used to build SMARTS strings
    for atom in offmol.atoms: 
        rep_res_nums[atom.metadata['residue_name']].add(atom.metadata['residue_number']) # collect unique residue numbers

    for res_name, ids in rep_res_nums.items():
        rep_res_nums[res_name] = min(ids) # choose group with smallest id of each residue to denote representative group
    LOGGER.info('Selected representative residue groups')

    return rep_res_nums

def compute_residue_charges(cmol : Molecule, monomer_group : MonomerGroup, cds : Optional[ChargeRedistributionStrategy]=UniformDistributionStrategy(desired_net_charge=0.0)) -> ChargesByResidue:
    '''
    Takes a charged molecule and a collection of monomer-spec SMARTS and generates monomer-averages library charges for each repeating residue
    Can optionally specify a strategy for redistributing any excess charge (by default, will ensure charges are neutral)

    Returns a ChargesByRresidue object containing a dict of mapped residue charges keyed by residue name
    '''
    rdmol = cmol.to_rdkit() # create rdkit representation of Molecule to allow for SMARTS generation
    rep_res_nums = find_repr_residues(cmol) # determine ids of representatives of each unique residue

    atom_id_mapping   = defaultdict(lambda : defaultdict(int))
    res_charge_accums = defaultdict(lambda : defaultdict(Accumulator))
    for atom in cmol.atoms: # accumulate counts and charge values across matching subsftructures
        res_name, res_num  = atom.metadata['residue_name'], atom.metadata['residue_number']
        substruct_id = atom.metadata['substructure_query_id']
        atom_id = atom.molecule_atom_index

        if res_num == rep_res_nums[res_name]: # if atom is member of representative group for any residue...
            # rdmol.GetAtomWithIdx(atom_id).SetAtomMapNum(atom_id)  # ...and set atom number for labelling in SMARTS string
            atom_id_mapping[res_name][atom_id] = (substruct_id, atom.symbol) # collect pdb id

        curr_accum = res_charge_accums[res_name][substruct_id] # accumulate charge info for averaging
        curr_accum.sum += atom.partial_charge.magnitude # eschew units (easier to handle, added back when writing to XML)
        curr_accum.count += 1
    LOGGER.info('Accumulated charges across all matching residues')

    chgs_by_res = ChargesByResidue()
    for res_name, charge_accums in res_charge_accums.items():
        SMARTS = monomer_group.monomers[res_name][0] # extract SMARTS string from monomer data
        # rdSMARTS = rdmolfiles.MolFragmentToSmarts(rdmol, atomsToUse=atom_id_mapping[res_name].keys()) # determine SMARTS for the current residue's representative group
        # mol_frag = rdmolfiles.MolFromSmarts(rdSMARTS) # create fragment from rdkit SMARTS to avoid wild atoms (using rdkit over nx.subgraph for more detailed atomwise info)

        charge_map = {substruct_id : accum.average for substruct_id, accum in charge_accums.items()} 
        if cds is not None:
            LOGGER.info(f'Redistributing charges for residue "{res_name}" according to {cds.__class__.__name__}')
            mol_frag = Chem.MolFromSmarts(SMARTS) # TODO : make this take specific fragments from the original molecule in an id-preserving way
            charge_map = cds.redistributed_charges(charge_map, mol_frag)

        chgs_by_res.charges[res_name] = charge_map
    LOGGER.info(f'Successfully computed library charges for Molecule "{cmol.name}"')

    return chgs_by_res

@optional_in_place
def apply_residue_charges(offmol : Molecule, chgs_by_res : ChargesByResidue) -> None:
    '''Takes an OpenFF Molecule and a residue-wise map of averaged partial charges and applies the mapped charges to the Molecule'''
    # TODO : add check for correctly residue-partitioned molecule by metadata
    new_charges = [
        chgs_by_res.charges[atom.metadata['residue_name']][atom.metadata['substructure_query_id']]
            for atom in offmol.atoms
    ]
    offmol.partial_charges = np.array(new_charges) * elementary_charge # convert to array with units (otherwise assignment won't work)
    LOGGER.info(f'Successfully mapped residue charges onto Molecule "{offmol.name}"')



'''For dynamically determining and cataloging which ToolkitWrappers (and accompanying functionality) are available'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

# Core OpenFF toolkit component registration
from typing import Union
from collections import defaultdict

from openff.toolkit.utils.utils import all_subclasses
from openff.toolkit.utils.base_wrapper import ToolkitWrapper
from openff.toolkit.utils.toolkit_registry import ToolkitRegistry 
from openff.toolkit.utils.toolkits import (
    OPENEYE_AVAILABLE,
    RDKIT_AVAILABLE,
    AMBERTOOLS_AVAILABLE,
    GLOBAL_TOOLKIT_REGISTRY,
)
GTR = GLOBAL_TOOLKIT_REGISTRY # alias for brevity

from ...genutils.importutils.dependencies import modules_installed
_REGISTER_TOOLKITS_TO_GLOBAL : bool = True # TODO: find way to avoid setting this config parameter directly in code


# Helper functions
def toolkit_wrapper_is_registered(toolkit_wrapper : Union[ToolkitWrapper, type[ToolkitWrapper]], toolkit_registry : ToolkitRegistry=GTR) -> bool:
    '''Check whether a ToolkitRegistry instance has already registered a given ToolkitWrapper subclass'''
    if not isinstance(toolkit_wrapper, type):   # ToolkitWrapper TYPES are needed for this check; any instances therefore...
        toolkit_wrapper = type(toolkit_wrapper) # ...will have their respective types extracted
    if not issubclass(toolkit_wrapper, ToolkitWrapper):
        raise TypeError(f'Expected a ToolkitWrapper instance or subclass, instead received object of type {toolkit_wrapper.__name__}')
        
    return any(isinstance(tkwrapper, toolkit_wrapper) for tkwrapper in toolkit_registry.registered_toolkits)

# Setup of initial containers for OpenFF module info
ALL_IMPORTABLE_TKWRAPPERS : list[type[ToolkitWrapper]] = all_subclasses(ToolkitWrapper) # NOTE: just because you can import these, doesn't necessarily mean they can be instantiated
ALL_AVAILABLE_TKWRAPPERS  : list[type[ToolkitWrapper]] = []
CHARGE_METHODS_BY_TOOLKIT : dict[type[ToolkitWrapper], list[str]] = defaultdict(list)

# Toolkit-specific registrations which depend on available packages
## BuiltIn (not particularly useful in and of itself, but nice to know it's accessible)
if modules_installed('openff.toolkit'): # this check is idempotent to initial OpenFF check, but is nice to have for consistency between all ToolkitWrappers below
    from openff.toolkit.utils.builtin_wrapper import BuiltInToolkitWrapper
    
    ALL_AVAILABLE_TKWRAPPERS.append(BuiltInToolkitWrapper)
    CHARGE_METHODS_BY_TOOLKIT[BuiltInToolkitWrapper] = [
        charge_method
            for charge_method in BuiltInToolkitWrapper._supported_charge_methods
    ]
## RDKit
if modules_installed('rdkit') and RDKIT_AVAILABLE:
    from openff.toolkit.utils.rdkit_wrapper import RDKitToolkitWrapper
    
    ALL_AVAILABLE_TKWRAPPERS.append(RDKitToolkitWrapper)
    CHARGE_METHODS_BY_TOOLKIT[RDKitToolkitWrapper] = [
        charge_method
            for charge_method in RDKitToolkitWrapper._supported_charge_methods
    ]
## Ambertools
if modules_installed('pdb4amber') and AMBERTOOLS_AVAILABLE: # turns out "ambertools" can't actually be imported as a module, need to check for peripheral modules which are better behaved instead
    from openff.toolkit.utils.ambertools_wrapper import AmberToolsToolkitWrapper
    
    ALL_AVAILABLE_TKWRAPPERS.append(AmberToolsToolkitWrapper)
    CHARGE_METHODS_BY_TOOLKIT[AmberToolsToolkitWrapper] = [
        charge_method
            for charge_method in AmberToolsToolkitWrapper._supported_charge_methods
    ]
## OpenEye
if modules_installed('openeye.oechem', 'openeye.oeomega') and OPENEYE_AVAILABLE:
    from openff.toolkit.utils.openeye_wrapper import OpenEyeToolkitWrapper
    
    ALL_AVAILABLE_TKWRAPPERS.append(OpenEyeToolkitWrapper)
    CHARGE_METHODS_BY_TOOLKIT[OpenEyeToolkitWrapper] = [
        charge_method
            for charge_method in OpenEyeToolkitWrapper._supported_charge_methods
    ]
## NAGL - extracting available charge methods is a little different for GNN toolkits
if modules_installed('openff.nagl', 'openff.nagl_models'):
    from openff.toolkit.utils.nagl_wrapper import NAGLToolkitWrapper
    
    ALL_AVAILABLE_TKWRAPPERS.append(NAGLToolkitWrapper)
    CHARGE_METHODS_BY_TOOLKIT[NAGLToolkitWrapper] = [
        model_path.name # need to extract dynamically from Paths
            for model_path in NAGLToolkitWrapper.list_available_nagl_models()
    ]
## Espaloma - extracting available charge methods is a little different for GNN toolkits
if modules_installed('espaloma_charge'):
    from espaloma_charge.openff_wrapper import EspalomaChargeToolkitWrapper
    
    ALL_AVAILABLE_TKWRAPPERS.append(EspalomaChargeToolkitWrapper)
    CHARGE_METHODS_BY_TOOLKIT[EspalomaChargeToolkitWrapper] = [
        'espaloma-am1bcc'
    ] # this is, at this of writing, the only available method for EspalomaCharge and unfortunately not accessible dynamically
    
# Post-registration info compilation
## Compiling name-based lookups for available ToolkitWrappers and registering all to a local
POLYMERIST_TOOLKIT_REGISTRY = ToolkitRegistry() # retain a local registry separate from GLOBAL_TOOLKIT_REGISTRY
TKWRAPPERS      : dict[str,      ToolkitWrapper ] = {} 
TKWRAPPER_TYPES : dict[str, type[ToolkitWrapper]] = {} 

for tkwrapper_type in ALL_AVAILABLE_TKWRAPPERS:
    tkwrapper_instance = tkwrapper_type() # instantiate toolkit wrapper class
    POLYMERIST_TOOLKIT_REGISTRY.register_toolkit(tkwrapper_instance)
    # if requested, also mirror all found toolkits to the Global ToolkitRegistry
    if _REGISTER_TOOLKITS_TO_GLOBAL and not toolkit_wrapper_is_registered(tkwrapper_type, GLOBAL_TOOLKIT_REGISTRY): # make registration idempotent
        GLOBAL_TOOLKIT_REGISTRY.register_toolkit(tkwrapper_instance)
    
    # register to name-based lookup dict
    TKWRAPPERS[     tkwrapper_type._toolkit_name] = tkwrapper_instance
    TKWRAPPER_TYPES[tkwrapper_type._toolkit_name] = tkwrapper_type

## Compiling registry of which partial charge methods are supported by which toolkits
TOOLKITS_BY_CHARGE_METHOD : dict[str, list[type[ToolkitWrapper]]] = defaultdict(list) # also compile inverse mapping (compiled once available toolkits are known)
for tkwrapper_type, supported_methods in CHARGE_METHODS_BY_TOOLKIT.items():
    if (tkwrapper_type in ALL_AVAILABLE_TKWRAPPERS): # exclude non-registered toolkits to avoid confusion
        for method in supported_methods:
            TOOLKITS_BY_CHARGE_METHOD[method].append(tkwrapper_type)
TOOLKITS_BY_CHARGE_METHOD = dict(TOOLKITS_BY_CHARGE_METHOD) # convert to pure dict for typing purposes



'''Tools for manipulating, processing, and pretty-printing text from files and string-like objects'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from string import ascii_letters, ascii_lowercase, ascii_uppercase
ascii_printable = ''.join(chr(i) for i in range(33, 127)) # ASCII printable characters, minus SPACE (" ", 32) and DELETE (127)


'''Utilities related to handling SMARTS queries'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Generator, TypeVar
T = TypeVar('T')

from rdkit import Chem
from rdkit.Chem import rdqueries, Mol


# REFERENCE TABLES FOR SPECIAL ATOM TYPES
_special = { # shorthand for special, non-element queries
    'metal'           : 'M',
    'halogen'         : 'X',
    'heavy'           : 'A',
    'heavy_noncarbon' : 'Q',
}

SPECIAL_SMARTS = { 
    query_label : getattr(rdqueries, f'{symbol}AtomQueryAtom')().GetSmarts()
        for query_label, symbol in _special.items()
}

SPECIAL_QUERY_MOLS = { # TODO : make these lambda-like so that a unique object is returned on access
    query_name : Chem.MolFromSmarts(smarts)
        for query_name, smarts in SPECIAL_SMARTS.items()
}


# COUNTING SUBSTRUCTURE QUERIES
def num_substruct_queries(target_mol : Mol, substruct_query : Mol, *args, **kwargs) -> int:
    '''Get the number of RDKit substruct matches to a SMARTS query within a given target Mol'''
    return len(target_mol.GetSubstructMatches(substruct_query, *args, **kwargs)) # default "asMols=False" is fine here for length

def num_automorphisms(substruct_query : Mol, *args, **kwargs) -> int:
    '''Get the matches a substructure query has to itself; provides measure of the degree of symmetry of the query'''
    kwargs['uniquify'] = False # force non-unique match to see how many possible ways the atoms in the substruct can be mapped to themselves
    return num_substruct_queries(substruct_query, substruct_query, *args, **kwargs)

def num_substruct_queries_distinct(target_mol : Mol, substruct_query : Mol) -> int: # TODO : also include stereochemical symmetries (via "useChirality" flag) 
    '''Get the number of distinct, non-overlapping RDKit substruct matches to a SMARTS query within a given target Mol
    Accounts for automorphic symmetries of the query to avoid double-counting distinct groups'''
    num_distinct : float = num_substruct_queries(target_mol, substruct_query, uniquify=False) # !CRITICAL! that matches be non-unique
    num_distinct /= num_automorphisms(substruct_query) # quotient out number of substructure automorphisms
    if num_distinct.is_integer(): # should be guaranteed to be an integer by Lagrange's Theorem, but good to double check here
        return int(num_distinct)
    else:
        raise ValueError('Automorphism normalization returned a non-integer number of query matches')


# MAPPING SUBSTRUCTURE QUERIES
def matching_labels_from_substruct_dict(target_mol : Mol, substruct_queries : dict[T, Mol]) -> Generator[T, None, None]:
    '''Takes a target RDKit Mol and a string-keyed dict of SMARTS substructure query Mols and 
    yields ONLY the keys of substructures which are found in the target'''
    for match_mol_name, match_mol in substruct_queries.items():
        if target_mol.HasSubstructMatch(match_mol):
            yield match_mol_name

def matching_dict_from_substruct_dict(target_mol : Mol, substruct_queries : dict[T, Mol]) -> dict[T, bool]:
    '''Takes a target RDKit Mol and a string-keyed dict of SMARTS substructure query Mols and 
    returns a dict of bools with the same keys indicating whether each match is present'''
    return {
        match_mol_name : target_mol.HasSubstructMatch(match_mol)
            for match_mol_name, match_mol in substruct_queries.items()
    }

def matching_dict_from_substruct_dict_alt(target_mol : Mol, substruct_queries : dict[T, Mol]) -> dict[T, bool]:
    '''Takes a target RDKit Mol and a string-keyed dict of SMARTS substructure query Mols and 
    returns a dict of bools with the same keys indicating whether each match is present
    
    Alternate implementation with lazy evaluation'''
    matching_dict = {
        match_mol_name : False # set False as sentinel value
            for match_mol_name in substruct_queries.keys()
    }

    for label in matching_labels_from_substruct_dict(target_mol, substruct_queries=substruct_queries):
        matching_dict[label] = True

    return matching_dict


'''Utilities for interfacing between RDKit Mols and their graph representations'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Iterable

from rdkit import Chem
import networkx as nx

from .rdprops import detailed_rdobj_info
from ..genutils.textual.casing import camel_case_to_snake_case


# CONVERTING Mols TO Graphs
DEFAULT_ATOM_PROPS = (
    'AtomMapNum',
    'AtomicNum',
    'Degree',
    'ExplicitValence',
    'FormalCharge',
    'ImplicitValence',
    'IsAromatic',
    'Isotope',
    'Mass',
    'NoImplicit',
    'NumExplicitHs',
    'NumImplicitHs',
    'NumRadicalElectrons',
    'Symbol',
    'TotalDegree',
    'TotalNumHs',
    'TotalValence',
)

DEFAULT_BOND_PROPS = (
    'BondDir',
    'BondType',
    'BondTypeAsDouble',
    'Idx',
    'IsAromatic',
    'IsConjugated',
    'Stereo',
)

def rdmol_to_networkx(rdmol : Chem.Mol, atom_attrs : Iterable[str]=DEFAULT_ATOM_PROPS, bond_attrs : Iterable[str]=DEFAULT_BOND_PROPS) -> nx.Graph:
    '''Convert an RDKit Mol into a NetworkX Graph with nodes numbered according to atom IDs and bonds between corresponding atoms 
    Copies over all Atom and Bond Props that have been set.
    Can also port other atom and bond properties over to the resulting graph by specifying attributes to copy over'''
    G = nx.Graph()
    for atom in rdmol.GetAtoms():
        atom_attr_vals = {
            camel_case_to_snake_case(attr) : attr_val
                for attr, attr_val in detailed_rdobj_info(atom).items()
                    if attr in atom_attrs
        }
        atom_attr_vals.update(atom.GetPropsAsDict())
        G.add_node(atom.GetIdx(), **atom_attr_vals)

    for bond in rdmol.GetBonds():
        bond_attr_vals = {
            camel_case_to_snake_case(attr) : attr_val
                for attr, attr_val in detailed_rdobj_info(bond).items()
                    if attr in bond_attrs
        }
        bond_attr_vals.update(atom.GetPropsAsDict())
        G.add_edge(bond.GetBeginAtomIdx(), bond.GetEndAtomIdx(), **bond_attr_vals)

    return G


'''Tools for breaking bonds in RDKit Mols and assigning new Ports'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import logging
LOGGER = logging.getLogger(__name__)

from typing import Optional
from rdkit import Chem
from rdkit.Chem.rdchem import BondType, RWMol

from ..rderrors import BondOrderModificationError
from ..labeling.bondwise import are_bonded_atoms
from ...genutils.decorators.functional import optional_in_place


@optional_in_place
def _decrease_bond_order(rwmol : RWMol, atom_id_1 : int, atom_id_2 : int) -> RWMol: 
    '''Lower the order of a bond between two atoms, raising Expection if no bond exists
    DOES NOT ENSURE VALENCE OF BONDED ATOMS IS PRESERVED'''
    if not are_bonded_atoms(rwmol, atom_id_1, atom_id_2):
        raise BondOrderModificationError
    
    # determine expected bond type after order decrease (handle single-bond case, specifically) 
    curr_bond = rwmol.GetBondBetweenAtoms(atom_id_1, atom_id_2) # guaranteed to not be None by the bond_order_decreasable check at the start
    new_bond_type = BondType.values[curr_bond.GetBondTypeAsDouble() - 1] # with pre-existing bond, need to get the next order up by numeric lookup
    if new_bond_type == BondType.UNSPECIFIED:
        new_bond_type = None # explicitly set to NoneType if single bond is broken

    # remove existing bond; not single bond, replace with bond of new type
    rwmol.RemoveBond(atom_id_1, atom_id_2) # create new bond or specified order
    if new_bond_type is not None:
        rwmol.AddBond(atom_id_1, atom_id_2, order=new_bond_type)

# SINGLE BOND-ORDER CHANGES 
@optional_in_place
def decrease_bond_order(rwmol : RWMol, atom_id_1 : int, atom_id_2 : int, new_flavor_pair : Optional[tuple[int, int]]=None) -> None: 
    '''Lower the order of a bond between two atoms and insert two new ports in its place, raising Exception if no bond exists'''
    atom_ids = (atom_id_1, atom_id_2)
    if new_flavor_pair is None:
        new_flavor_pair = (0, 0)
    assert(len(new_flavor_pair) == 2)
    
    _decrease_bond_order(rwmol, atom_id_1, atom_id_2, in_place=True) # NOTE : must explicitly be called in-place to ensure correct top-level behavior, since this function is also decorated
    # free_isotope_labels = int_complement(get_isotopes(rwmol, unique=True), bounded=False) # generate unused isotope labels
    # add new ports for broken bond
    for atom_id, flavor in zip(atom_ids, new_flavor_pair):
        new_linker = Chem.AtomFromSmarts('[#0]') # TODO : see if this can be made into a SMILES mol instead?
        new_linker.SetIsotope(flavor)
        new_port_id = rwmol.AddAtom(new_linker)# insert new port into molecule, taking note of index (TOSELF : ensure that this inserts indices at END of existing ones, could cause unexpected modification if not)
        
        rwmol.AddBond(atom_id, new_port_id, order=BondType.SINGLE) # bond the atom to the new port
        # _increase_bond_order(rwmol, atom_id, new_port_id)

@optional_in_place
def dissolve_bond(rwmol : RWMol, atom_id_1 : int, atom_id_2 : int, new_flavor_pair : Optional[tuple[int, int]]=None) -> None:
    '''Completely decompose a bond between two atoms, filling in ports with the chosen flavor''' 
    while rwmol.GetBondBetweenAtoms(atom_id_1, atom_id_2) is not None:
        decrease_bond_order(rwmol, atom_id_1, atom_id_2, new_flavor_pair=new_flavor_pair, in_place=True)



'''Tools for simplifying iteration over collections of items'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Any, Callable, Generator, Iterable, TypeVar, Union

from operator import mul
from functools import reduce
from collections import deque
from itertools import islice, combinations, product as cartesian_product

from .typetools.parametric import Args, KWArgs
from .decorators.functional import optional_in_place


# READ-ONLY ITERATION
T = TypeVar('T') # generic type of object(s) in iterator
T1 = TypeVar('T1') # generic type of object(s) in first iterator
T2 = TypeVar('T2') # generic type of object(s) in second iterator


def iter_len(itera : Iterable):
    '''Get size of an iterable object where ordinary len() call is invalid (namely a generator)
    Note that this will "use up" a generator upon iteration'''
    return sum(1 for _  in itera)

def product(itera : Iterable[T]) -> T:
    '''Multiplicative analogue to builtin sum()'''
    return reduce(mul, itera) # assumes that the type T implements __mul__/__rmul__

def sliding_window(items : Iterable[T], n : int=1) -> Generator[tuple[T], None, None]:
    '''Generates sliding windows of width n over an iterable collection of items
    E.g. : sliding_window('ABCDE', 3) --> (A, B, C), (B, C, D), (C, D, E)
    '''
    it = iter(items)
    window = deque(islice(it, n), maxlen=n)
    if len(window) == n:
        yield tuple(window)

    for x in it: # implicit else
        window.append(x)
        yield tuple(window)

def subsets(items : Iterable[T], exclude_empty : bool=False, exclude_full : bool=False) -> Generator[tuple[T], None, None]:
    '''Generate all possible subsets of a set. Can optionally exclude the empty set or the complete set (or both)'''
    base_set = set(items)
    for i in range(int(exclude_empty), len(base_set) + int(not exclude_full)):
        yield from combinations(base_set, i)

def swappable_loop_order(iter1 : Iterable[T1], iter2 : Iterable[T2], swap : bool=False) -> Union[Iterable[tuple[T1, T2]], Iterable[tuple[T2, T1]]]:
    '''Enables dynamic swapping of the order of execution of a 2-nested for loop'''
    order = [iter1, iter2] if not swap else [iter2, iter1]
    for pair in cartesian_product(*order):
        yield pair[::(-1)**swap] # reverse order of pair (preserves argument identity)

def progress_iter(itera : Iterable[T], key : Callable[[T], str]=lambda x : x) -> Iterable[tuple[str, T]]:
    '''Iterate through'''
    N = len(itera) # TODO : extend this to work for generators / consumables
    for i, item in enumerate(itera):
        yield (f'{key(item)} ({i + 1} / {N})', item) # +1 converts to more human-readable 1-index for step count


# READ-WRITE ITERATION (MODIFIES THE OBJECTS BEING ITERATED OVER)
K = TypeVar('K') # generic type for dict key
V = TypeVar('V') # generic type for dict value
O = TypeVar('O') # generic type for an object passed to a function

def asiterable(arg_val : Union[O, Iterable[O]]) -> Iterable[O]:
	'''Permits functions expecting iterable arguments to accept singular values'''
	if not isinstance(arg_val, Iterable):
		arg_val = (arg_val,) # turn into single-item tuple (better for memory)
	return arg_val

@optional_in_place
def modify_dict(path_dict : dict[K, V], modifier_fn : Callable[[K, V, Args, KWArgs], Any]) -> None:
    '''Recursively modifies all values in a dict in-place according to some function'''
    for key, val in path_dict.items():
        if isinstance(val, dict): # recursive call if sub-values are also dicts with Paths
            modify_dict(val, modifier_fn)
        else:
            path_dict[key] = modifier_fn(key, val) 

def sort_dict_by_keys(targ_dict : dict, reverse : bool=False) -> dict[Any, Any]:
    '''Sort a dictionary according to the values of each key'''
    return { # sort dict in ascending order by size
        key : targ_dict[key]
            for key in sorted(targ_dict, reverse=reverse)
    }

def sort_dict_by_values(targ_dict : dict, reverse : bool=False) -> dict[Any, Any]:
    '''Sort a dictionary according to the values of each key'''
    return { # sort dict in ascending order by size
        key : targ_dict[key]
            for key in sorted(targ_dict, key=lambda k : targ_dict[k], reverse=reverse)
    }


'''Tests construction of structures for linear copolymers (and relevant subfamilies, e.g. homopolymers)'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import pytest

from collections import Counter

from polymerist.polymers.monomers.repr import MonomerGroup
from polymerist.polymers.monomers.fragments import PE_FRAGMENTS, MPD_TMC_FRAGMENTS, PEG_PLGA_FRAGMENTS

from polymerist.polymers.building import build_linear_polymer
from polymerist.polymers.exceptions import MorphologyError, PartialBlockSequence, EmptyBlockSequence


@pytest.fixture(scope='function')
def monogrp_polyethylene() ->  MonomerGroup:
    return MonomerGroup(monomers=PE_FRAGMENTS)

@pytest.fixture(scope='function')
def monogrp_mpd_tmc() ->  MonomerGroup:
    return MonomerGroup(monomers=MPD_TMC_FRAGMENTS)

@pytest.fixture(scope='function')
def monogrp_peg_plga() ->  MonomerGroup:
    return MonomerGroup(monomers=PEG_PLGA_FRAGMENTS)


@pytest.mark.parametrize(
    'monomers, term_orient, n_monomers, sequence, minimize_sequence, allow_partial_sequences, energy_minimize',
    [
        # Polyethylene
        ('monogrp_polyethylene', {}, 7, 'A', True, True, False), # test end group autogen (should only have 1 term group)
        ('monogrp_polyethylene', {'head':'PE1', 'tail' : 'PE1'}, 7, 'A', True, True, False), # test explicit head-tail (result here should be different from autogen structure)
        ('monogrp_polyethylene', {'head':'PE1', 'tail' : 'PE1'}, 7, 'A', True, False, False), # test partial sequences (irrelevant here)
        ('monogrp_polyethylene', {'head':'PE1', 'tail' : 'PE1'}, 7, 'A', True, False, True), # test energy minimization doesn't crash
        pytest.param(  # will fail due to too few monomers for given sequence - 
            'monogrp_polyethylene', {}, 7, '', True, True, False, # NOTE: need to have partials enabled, since failure happens ONLY once sequence is passed to mbuild
            marks=pytest.mark.xfail(
                raises=EmptyBlockSequence,
                reason='Sequence provided must be nonempty',
                strict=True,
            )
        ),
        pytest.param(  # will fail due to too few monomers for given sequence - 
            'monogrp_polyethylene', {'head':'PE1', 'tail' : 'PE1'}, 7, 'AB', True, True, False, # NOTE: need to have partials enabled, since failure happens ONLY once sequence is passed to mbuild
            marks=pytest.mark.xfail(
                raises=ValueError,
                reason='Fewer unique monomers defined than called for by target sequence',
                strict=True,
            )
        ),
        # MPD-TMC
        ('monogrp_mpd_tmc', {'head':'MPD-1', 'tail' : 'TMC-1'}, 8, 'A', True, True, False), # correctly-specified: explicit end groups, only linear middle monomers, and whole number of sequence repeats
        pytest.param(
            'monogrp_mpd_tmc', {'head':'MPD-1', 'tail' : 'TMC-1'}, 7, 'AB', True, False, False, # will fail due to partial sequence
            marks=pytest.mark.xfail(
                raises=PartialBlockSequence,
                reason='Partial sequence repeat needed to get odd number block out of AB, but partial blocks are disabled',
                strict=True,
            )
        ),
        pytest.param(
            'monogrp_mpd_tmc', {'head':'MPD-1', 'tail' : 'TMC-1'}, 12, 'ABC', True, True, False, # will fail due to 3-functional TMC middle monomer as C
            marks=pytest.mark.xfail(
                raises=MorphologyError,
                reason='One of the monomers requested is non-linear (3-functional)',
                strict=True,
            )
        ),
        # PEG-PLGA
        ('monogrp_peg_plga', {}, 15, 'ABC', True, True, False), # test autogen
        ('monogrp_peg_plga', {}, 17, 'ABC', True, False, False), # test autogen with whole sequence
        ('monogrp_peg_plga', {'head':'PGA-1A', 'tail' : 'PGA-1B'}, 15, 'ABC', True, True, False), # test more complex sequence with non-default explicit end groups
        pytest.param(
            'monogrp_peg_plga', {'head':'PGA-1A', 'tail' : 'PGA-1B'}, 15, 'ABC', True, False, False, # will fail due to partial sequence
            marks=pytest.mark.xfail(
                raises=PartialBlockSequence,
                reason='Partial sequence repeat needed to get odd number block out of AB, but partial blocks are disabled',
                strict=True,
            )
        ),
        ('monogrp_peg_plga', {}, 40, 'ABCB', True, True, True), # test longer energy min
    ]
)
def test_build_linear_polymer(
        monomers : MonomerGroup,
        term_orient : dict[str, str],
        n_monomers : int,
        sequence : str,
        minimize_sequence : bool,
        allow_partial_sequences : bool,
        energy_minimize : bool,
        request : pytest.FixtureRequest, # allows for fixture expansion in parameterized arguments
    ) -> None:
    '''Test linear polymer builder behavior under varing sets of parameters'''
    monomers = request.getfixturevalue(monomers) # unpack fixtures into their respective values
    monomers.term_orient = term_orient # this edit makes it VITAL that fixtures be function-level
    
    polymer = build_linear_polymer(
        monomers=monomers,
        n_monomers=n_monomers,
        sequence=sequence,
        minimize_sequence=minimize_sequence,
        allow_partial_sequences=allow_partial_sequences,
        add_Hs=False,
        energy_minimize=energy_minimize,
    )
    
    # characterize middle monomers
    n_rep_units = len(polymer.children)
    residue_sizes : dict[str, int] = {}
    residue_counts = Counter() # TODO: make use of this for checks!!
    for middle_monomers in polymer.children:
        residue_sizes[middle_monomers.name] = middle_monomers.n_particles
        residue_counts[middle_monomers.name] += 1
        
    # characterize end groups
    end_groups_requested = set(resname for head_or_tail, (resname, mol) in monomers.linear_end_groups().items())
    end_groups_used = set()
    for end_group in polymer.end_groups:
        if end_group is not None:
            end_groups_used.add(end_group.name)
            residue_sizes[end_group.name] = end_group.n_particles
            residue_counts[middle_monomers.name] += 1
    
    total_reps_match = (n_rep_units == n_monomers)
    contribs_match = all(num_monomers == monomers.contributions()[resname][0]
        for resname, num_monomers in residue_sizes.items()
    )
    end_groups_correct = (end_groups_used == end_groups_requested)
    # counts_match = ...
    
    assert all([total_reps_match, contribs_match, end_groups_correct]) #, and counts_match    )


'''For gathering information and running calculations from LAMMPS input files'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Optional, Union

import re
from pathlib import Path

import lammps
from openmm.unit import Unit, Quantity
from openmm.unit import degree, kilocalorie_per_mole

from .unitstyles import LAMMPSUnitStyle
from ...genutils.decorators.functional import allow_string_paths, allow_pathlib_paths


_DISTANCE_WILDCARD : str = '$DISTANCE'
LMP_UNIT_REGEX = re.compile(r'^units\s(?P<unit_style>\w*)$')
LMP_THERMO_STYLE_REGEX = re.compile(r'^thermo_style\s(?P<thermo_style>\b\w*?\b)\s(?P<energy_evals>.*)$')

LAMMPS_ENERGY_KW : dict[str, str] = { # keywordS for evaluating energies and their corresponding meanings
    'ebond'  : 'Bond',
    'eangle' : 'Angle',
    'edihed' : 'Proper Torsion',
    'eimp'   : 'Improper Torsion',
    'ecoul'  : 'Coulomb Short',
    'elong'  : 'Coulomb Long',
    'evdwl'  : 'vdW',
    'etail'  : 'Dispersion',
    'epair'  : 'Nonbonded',
    'pe'     : 'Potential',
    'ke'     : 'Kinetic',
    'etotal' : 'Total'
}

LAMMPS_CELL_KW = ( # keywords for probing unit cell sizes and angles
    'cella', # TODO : add aliases
    'cellb',
    'cellc',
    'cellalpha',
    'cellbeta',
    'cellgamma',
)

LAMMPS_CELL_UNITS : dict[str, Union[str, Unit]] = { # units associated with unit cell keywords
    'cella' : _DISTANCE_WILDCARD, 
    'cellb' : _DISTANCE_WILDCARD,
    'cellc' : _DISTANCE_WILDCARD,
    'cellalpha' : degree,
    'cellbeta'  : degree,
    'cellgamma' : degree,
}

@allow_string_paths
def parse_lammps_input(lmp_in_path : Path) -> dict[str, Union[str, LAMMPSUnitStyle]]: # NOTE : this can and will be expanded in the future
    '''Read unit style, thermo style, and energies to evaluate from a LAMMPS input file'''
    info_dict = {}
    with lmp_in_path.open('r') as lmp_in_file:
        for line in lmp_in_file.read().split('\n'):
            if (thermo_match := re.search(LMP_THERMO_STYLE_REGEX, line)):
                info_dict.update(thermo_match.groupdict())
                info_dict['energy_evals'] = info_dict['energy_evals'].split(' ') # separate on spaces (TODO : maybe find more elegant way to do this in the future?)
            
            if (units_match := re.search(LMP_UNIT_REGEX, line)):
                unit_style_name = units_match.groupdict()['unit_style']
                try:
                    info_dict['unit_style'] = LAMMPSUnitStyle.subclass_registry[unit_style_name]
                except KeyError:
                    raise ValueError(f'Invalid unit style "{unit_style_name}" specified in LAMMPS input file')
                
    return info_dict

@allow_pathlib_paths
def get_lammps_energies(lmp_in_path : str, preferred_unit : Unit=kilocalorie_per_mole, cmdargs : Optional[list]=None) -> dict[str, Quantity]:
    '''Perform an energy evaluation using a LAMMPS input file
    Alternative to interchange.drivers.get_lammps_energies which is dynamically aware of energy units and assumes nothing about which energies are specified by thermo_style'''
    cmdargs = [] if cmdargs is None else [arg for arg in cmdargs] # need to copy list to keep read-only (lammps.lammps() modifies the argument list passed in-place)
    
    assert(preferred_unit.is_compatible(kilocalorie_per_mole)) # whatever unit is desired, it must be one of energy
    lammps_info = parse_lammps_input(lmp_in_path)
    energy_unit     = lammps_info['unit_style'].energy
    energy_contribs = lammps_info['energy_evals']

    with lammps.lammps(cmdargs=cmdargs) as lmp: # need to create new lammps() object instance for each run
        # lmp.commands_string( ENERGY_EVAL_STR.replace('$INP_FILE', str(lammps_file)) )
        lmp.file(lmp_in_path) # read input file and calculate energies; NOTE that this NEEDS to be a string (not Path!)
        return {
            f'{LAMMPS_ENERGY_KW[contrib]}' : (lmp.get_thermo(contrib) * energy_unit).in_units_of(preferred_unit)
                for contrib in energy_contribs
        }

@allow_pathlib_paths
def get_lammps_unit_cell(lmp_in_path : str, cmdargs : Optional[list]=None) -> dict[str, Quantity]: # TODO: reimplement this with maths.lattices.bravais.LatticeParameters
    '''Extract the 6 unit cell parameters specified by a LAMMPS input file'''
    cmdargs = [] if cmdargs is None else [arg for arg in cmdargs] # need to copy list to keep read-only (lammps.lammps() modifies the argument list passed in-place)
    
    lammps_info = parse_lammps_input(lmp_in_path)
    length_unit = lammps_info['unit_style'].distance

    with lammps.lammps(cmdargs=cmdargs) as lmp: # need to create new lammps() object instance for each run
        lmp.file(lmp_in_path) # read input file and calculate energies; NOTE that this NEEDS to be a string (not Path!)
        cell_params = {}
        for cell_param_kw in LAMMPS_CELL_KW:
            cell_param_unit = LAMMPS_CELL_UNITS[cell_param_kw]
            if cell_param_unit == _DISTANCE_WILDCARD:
                cell_param_unit = length_unit

            cell_params[cell_param_kw] = lmp.get_thermo(cell_param_kw) * cell_param_unit
    
    return cell_params


'''For reading and writing OpenMM components to files'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import logging
LOGGER = logging.getLogger(__name__)

from typing import Any, Optional, Union
from numpy.typing import NDArray
from dataclasses import dataclass

from pathlib import Path
from collections import Counter

from openmm import System, Context, State
from openmm import XmlSerializer, Vec3
from openmm.app import Simulation, PDBFile
from openmm.app import Topology as OpenMMTopology

from .parameters import SimulationParameters
from ...genutils.decorators.functional import allow_string_paths
from ...genutils.fileutils.pathutils import assemble_path
from ...genutils.fileutils.jsonio.jsonify import make_jsonifiable
from ...genutils.fileutils.jsonio.serialize import PathSerializer
from ...molfiles.pdb import SerialAtomLabeller


# DEFINING AND STORING SIMULATION PATHS
@make_jsonifiable(type_serializer=PathSerializer)
@dataclass
class SimulationPaths:
    '''Encapsulates Paths to various files associated with an OpenMM Simulation'''
    parameters_path   : Optional[Path] = None
    paths_path        : Optional[Path] = None

    system_path       : Optional[Path] = None
    topology_path     : Optional[Path] = None
    state_path        : Optional[Path] = None
    checkpoint_path   : Optional[Path] = None
    trajectory_path   : Optional[Path] = None
    
    state_data_path   : Optional[Path] = None
    time_data_path    : Optional[Path] = None
    spatial_data_path : Optional[Path] = None

    @allow_string_paths
    def init_top_and_sys_paths(self, out_dir : Path, prefix : str, record : bool=True) -> tuple[Path, Path]:
        '''Initialize Topology and System output paths for a given directory'''
        topology_path = assemble_path(out_dir, prefix, extension='pdb', postfix='topology')
        system_path   = assemble_path(out_dir, prefix, extension='xml', postfix='system')
        state_path    = assemble_path(out_dir, prefix, extension='xml', postfix='state')

        if record:
            self.topology_path = topology_path
            self.system_path = system_path
            self.state_path = state_path

        return topology_path, system_path, state_path
    
    @classmethod
    def from_dir_and_parameters(cls, out_dir : Path, prefix : str, sim_params : SimulationParameters, touch : bool=True) -> 'SimulationPaths':
        '''Create file directory and initialize simulationPaths object from a set of SimulationParameters'''
        path_obj = cls() # create empty path instance

        path_obj.parameters_path = assemble_path(out_dir, prefix, extension='json', postfix='parameters')
        path_obj.init_top_and_sys_paths(out_dir, prefix, record=True)  # initialize simulation saving files

        sim_params.reporter_params.init_reporter_paths(out_dir, prefix) # initialize simulation reporter files 
        for rep_label, path in sim_params.reporter_params.reporter_paths.items():
            setattr(path_obj, f'{rep_label}_path', path) # register reporter output paths; NOTE :relies on ReporterConfigs having matching labels

        if touch: # ensure files exist 
            out_dir.mkdir(exist_ok=True) # ensure the directory exists before attempting output
            sim_params.to_file(path_obj.parameters_path) # record simulation parameters
        
            sim_paths_path = assemble_path(out_dir, prefix, extension='json', postfix='paths')
            path_obj.paths_path = sim_paths_path # record path the SimulationsPaths object is being stored at within the object itself
            path_obj.to_file(sim_paths_path)
            
            for val in path_obj.__dict__.values():
                if isinstance(val, Path):
                    val.touch()

        return path_obj


# SERIALIZATION AND DESERIALIZATION FUNCTIONS # TODO : add assertions for file extensions
DEFAULT_STATE_PROPS : dict[str, bool] = {
    'getPositions'  : True,
    'getVelocities' : True,
    'getForces'     : True,
    'getEnergy'     : True,
    'getParameters' : True,
    'getParameterDerivatives' : False,
    'getIntegratorParameters' : False
}

@allow_string_paths
def serialize_state_from_context(state_path : Path, context : Context, state_params : dict[str, bool]=DEFAULT_STATE_PROPS) -> None:
    '''For saving State data within an existing OpenMM Context to file'''
    state = context.getState(**state_params)
    with state_path.open('w') as file:
        file.write(XmlSerializer.serialize(state))

def apply_state_to_context(context : Context, state : State) -> None: # TOSELF : this might be replaced with Context.getState() followed by context.reinitialize(preserveState=True)
    '''For applying saved State data to an existing OpenMM Context'''
    context.setPeriodicBoxVectors(*state.getPeriodicBoxVectors())
    context.setPositions(state.getPositions())
    context.setVelocities(state.getVelocities())
    context.setTime(state.getTime())

    context.reinitialize(preserveState=True)


@allow_string_paths
def serialize_system(sys_path : Path, system : System) -> None:
    '''For saving an existing OpenMM System to file'''
    with sys_path.open('w') as file:
        file.write(XmlSerializer.serialize(system))

@allow_string_paths
def serialize_openmm_pdb(
        pdb_path : Path,
        topology : OpenMMTopology,
        positions : Union[NDArray, list[Vec3]],
        keep_chain_and_res_ids : bool=True,
        atom_labeller : Optional[SerialAtomLabeller]=SerialAtomLabeller(),
        resname_map : Optional[dict[str, str]]=None,
    ) -> None:
    '''Configure and write an Protein DataBank File from an OpenMM Topology and array of positions
    Provides options to configure atom ID numbering, residue numbering, and residue naming'''
    if resname_map is None:
        resname_map = {} # avoids mutable default

    # chain config
    for chain in topology.chains():
        chain.id = str(chain.id)

    # residue config
    for residue in topology.residues():
        residue.id = str(residue.id) # avoids TypeError when specifying keepIds during PDB write
        repl_res_name = resname_map.get(residue.name, None) # lookup current residue name to see if a replacement is called for
        if repl_res_name is not None:
            residue.name = repl_res_name

    # individual atom config
    if atom_labeller: # implicitly, preserves extant atom names if a labeller is not given
        for atom in topology.atoms():
            atom.name = atom_labeller.get_atom_label(atom.element.symbol)

    # file write
    with pdb_path.open('w') as file:
        PDBFile.writeFile(topology, positions, file, keepIds=keep_chain_and_res_ids)

@allow_string_paths
def serialize_topology_from_simulation(pdb_path : Path, sim : Simulation, keep_ids : bool=False) -> None:
    '''Saves a PDB of the current state of a simulation's Topology'''
    serialize_openmm_pdb(
        pdb_path,
        topology=sim.topology,
        positions=sim.context.getState(getPositions=True).getPositions(),
        keep_chain_and_res_ids=keep_ids
    )


'''
Tools for solving the DISCERNMENT (Determination of Index Sequences from Complete Enumeration of Ransom Notes - Multiset Extension with Nonlexical Types) problem

DISCERNMENT problem definition:
    Given a "word" (a sequence of N symbols of type T), and a mapped sequence of "bins" (ordered collection of multisets of type T, each assigned a label of type L),
    enumerate all N-tuples of labels such that the symbols of the words could be drawn from the bins with those labels in that order
'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from .enumeration import DISCERNMENTSolver
from .strategies import (
    DISCERNMENTStrategyStack,
    DISCERNMENTStrategyCartesian,
    DISCERNMENTStrategyRecursive,
)


'''Boilerplate for setting up OpenMM Simulations and related files'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import logging
LOGGER = logging.getLogger(__name__)

from typing import Optional, Union
from pathlib import Path

from openmm import State, System, XmlSerializer
from openmm.app import Simulation, Topology
from openmm.unit import Quantity

# from .records import SimulationParameters
from .serialization import SimulationPaths
from .thermo import EnsembleFactory
from .parameters import ThermoParameters, SimulationParameters
from .forcegroups import impose_unique_force_groups


def load_state_flexible(state : Optional[Union[str, Path, State]]=None) -> Optional[State]:
    '''Allows one to flexibly load an OpenMM state, either from a State object or file-like object'''
    if isinstance(state, State) or (state is None):
        state = state
    else:
        if isinstance(state, Path):
            state_path = state
        elif isinstance(state, str):
            state_path = Path(state)
        # TODO : add support for load from opened file
        else:
            raise TypeError('State can only be loaded from pathlike object') 
        
        try:
            with state_path.open('r') as state_file:
                state = XmlSerializer.deserialize(state_file.read())
        except ValueError:
            state = None
    
    if state is None:
        LOGGER.warning('No valid State/State file provided, initializing State as None')
    return state

def simulation_from_thermo(topology : Topology, system : System, thermo_params : ThermoParameters, time_step : Quantity, state : Optional[Union[str, Path, State]]=None) -> Simulation:
    '''Prepare an OpenMM simulation from a serialized thermodynamics parameter set'''
    ens_fac = EnsembleFactory.from_thermo_params(thermo_params)
    if (extra_forces := ens_fac.forces()): # check if any extra forces are present
        for force in extra_forces:
            system.addForce(force) # add forces to System BEFORE creating Simulation to avoid having to reinitialize the Conext to preserve changes 
            LOGGER.info(f'Added {force.getName()} Force to System')
    impose_unique_force_groups(system)

    simulation = Simulation(
        topology=topology,
        system=system,
        integrator=ens_fac.integrator(time_step),
    )
    state = load_state_flexible(state)
    if state is not None:
        LOGGER.info('Setting simulation state')
        simulation.context.setState(state)
        simulation.context.reinitialize(preserveState=True) # TOSELF : unclear whether this is necessary, redundant, or in fact harmful

    return simulation

def initialize_simulation_and_files(out_dir : Path, prefix : str, sim_params : SimulationParameters, topology : Topology, system : System, positions : Optional[Quantity]=None) -> tuple[Simulation, SimulationPaths]:
    '''Create simulation, bind Reporters, and update simulation Paths with newly-generated files'''
    sim_paths = SimulationPaths.from_dir_and_parameters(out_dir, prefix, sim_params, touch=True)

    # create simulation and add reporters
    simulation = simulation_from_thermo(topology, system, sim_params.thermo_params, time_step=sim_params.integ_params.time_step, state=sim_paths.state_path)
    for reporter in sim_params.reporter_params.prepare_reporters(report_interval=sim_params.integ_params.report_interval):
        simulation.reporters.append(reporter) # add reporters to simulation instance

    if positions is not None:
        # TODO : add position type/shape checking
        LOGGER.info('Setting positions in Context')
        simulation.context.setPositions(positions) # set positions if provided

    return simulation, sim_paths


'''Generic functionality for tree-like data structures. Based on the anytree module (https://github.com/c0fec0de/anytree)'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from .treebase import NodeCorrespondence, compile_tree_factory
from .treecopy import get_node_attrs, copy_tree, tree_to_networkx
from .treeviz import treestr


'''Tools for statically or dynamically updating JSON files'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import json
from pathlib import Path

from .serialize import JSONSerializable
from ...decorators.functional import allow_string_paths


@allow_string_paths
def append_to_json(json_path : Path, **kwargs) -> None:
    '''Add an entry to an existing JSON file'''
    with json_path.open('r') as json_file:
        jdat = json.load(json_file)

    jdat.update(**kwargs)

    with json_path.open('w') as json_file:
        jdat = json.dump(jdat, json_file, indent=4)

class JSONDict(dict): # TODO : add allow_str_path decorators
    '''Dict subclass which also updates an underlying JSON file - effectively and on-disc dict
    !NOTE! - JSON doesn't support non-string keys, so all keys given will be stringified - plan accordingly!'''
    def __init__(self, json_path : Path, *args, **kwargs):
        if isinstance(json_path, str):
            json_path = Path(json_path) # make input arg a bit more flexible to str input from user end

        if json_path.suffix != '.json':
            raise ValueError(f'The path "{json_path}" does not point to a .json file')
        
        self.json_path : Path = json_path
        if self.json_path.exists():
            try:
                kwargs.update(self._read_file(json_path))
            except json.JSONDecodeError: # catches Paths which point to incorrectly formatted JSONs - TODO: revise terrible except-pass structure
                pass
        else:
            self.json_path.touch()

        super().__init__(*args, **kwargs)
        self._update_file() # ensure file contains current contents post-init

    @staticmethod
    def _read_file(json_path : Path) -> dict:
        with json_path.open('r') as file:
            return json.load(file)        

    def _update_file(self, indent : int=4):
        '''Save current dict contents to JSON file'''
        with self.json_path.open('w') as file:
            json.dump(self, file, indent=indent)

    def __setitem__(self, __key: str, __value: JSONSerializable) -> None:
        super().__setitem__(__key, __value)
        self._update_file()

    def __delitem__(self, __key: str) -> None:
        super().__delitem__(__key)
        self._update_file()


'''Interfaces for encoding arbitrary classes into tree-like data structures'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Any, Callable, Generic, Iterable, Optional, TypeAlias, TypeVar
from abc import ABC, abstractmethod

from anytree.node import Node

from ..decorators.classmod import register_abstract_class_attrs
from ..filters import Filter, ALWAYS_FALSE_FILTER

T = TypeVar('T')


@register_abstract_class_attrs('FROMTYPE') # TODO: figure out way to parameterize Generic T here with the type passed as FROMTYPE
class NodeCorrespondence(ABC, Generic[T]): 
    '''Abstract base for implementing how to build an anytree Node tree for an arbitrary class'''
    @abstractmethod
    def name(self, obj : T) -> str:
        '''Define how to obtain a string name'''
        pass

    @abstractmethod
    def has_children(self, obj : T) -> bool:
        '''Define how to check if an object can produce children in the first place before attempting to do so'''
        pass

    @abstractmethod
    def children(self, obj : T) -> Optional[Iterable[T]]:
        '''Define how to obtain node children from an instance
        Should return NoneType if the instance is "leaf-like"'''
        pass

def compile_tree_factory(
        node_corresp : NodeCorrespondence[T],
        class_alias : Optional[str]=None,
        obj_attr_name : Optional[str]=None,
        exclude_mixin : Optional[Filter[T]]=None,
    ) -> Callable[[T, Optional[int], Optional[Filter[T]]], Node]:
    '''
    Factory method for producing a tree-generating function from a NodeCorrespondence
    
    Parameters
    ----------
    node_corresp : NodeCorrespondence[T]
        Definition of a correpondence between an arbitrary type and a Tree Node
    class_alias : str (optional)
        Name of the corresponding class to inject into docstring
        If not provided, will default to the __name__ of the class wrapped by node_corresp
    obj_attr_name : str (optional)
        The name of the Node attribute to which a copy of the class instance should be bound
        If not provided, will default to the value of class_alias
    exclude_mixin : Filter[T] (optional)
        An optional "master" filter to mix into any 

    Returns
    -------
    compile_tree : Callable[[T, Optional[int], Optional[Filter[T]]], Node]
        Factory function which takes an instance of type T, builds a Tree from it, and returns the root Node
    '''
    node_type_name = node_corresp.FROMTYPE.__name__
    if class_alias is None: # an alternative name to use when describing the tree creation for this class
        class_alias = node_type_name

    if obj_attr_name is None: # the name given to the Node attribute which store an instance of the given arbitrary type
        obj_attr_name = class_alias
    if hasattr(Node, obj_attr_name):
        raise AttributeError(f'Invalid value for obj_attr_name; attribute "{obj_attr_name}" clashes with existing attribute Node.{obj_attr_name}')

    if exclude_mixin is None:
        exclude_mixin = ALWAYS_FALSE_FILTER

    def compile_tree(
        obj : node_corresp.FROMTYPE,
        max_depth : Optional[int]=None,
        exclude : Filter[node_corresp.FROMTYPE]=ALWAYS_FALSE_FILTER,
        _curr_depth : int=0
    ) -> Node:
        # NOTE: deliberately omitting docstring here, as it will be built procedurally after defining this function
        node = Node(name=node_corresp.name(obj))
        setattr(node, obj_attr_name, obj) # keep an instance of the object directly for reference

        if node_corresp.has_children(obj) and ( # recursively add subnodes IFF
                (max_depth is None)             # 1) no depth limit is set, or
                or (_curr_depth < max_depth)    # 2) a limit IS set, but hasn't been reached yet
            ): 
            for child_obj in node_corresp.children(obj):
                if not (exclude(child_obj) or exclude_mixin(child_obj)):
                    sub_node = compile_tree(child_obj, max_depth=max_depth, exclude=exclude, _curr_depth=_curr_depth+1)
                    sub_node.parent = node

        return node

    # annoyingly, docstrings must be string literals (this CAN'T be done inside the function definition)
    compile_tree.__doc__ = f''' 
    Compile a {class_alias} tree from a(n) {node_type_name} object

    Any sub-{class_alias} encountered will be expanded into its own tree,
    up to the specified maximum depth, or until exhaustion if max_depth=None

    Parameters
    ----------
    obj : {node_type_name}
        A(n) instance of a {class_alias}
    max_depth : int (optional)
        Maximum allowed height of a constructed tree from the root
        If None (as default), no limit is set
    exclude : Filter[{node_type_name}] (optional)
        An optional filter function to reduce the size of the constructed tree
        Must accept a(n) {node_type_name} instance as single argument
        Should return True when a node is to be excluded, and False otherwise
    '''

    return compile_tree


'''For estimating properties of chains based on their constituent monomers and chain info'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import numpy as np

from .exceptions import InsufficientChainLength
from ..polymers.monomers.repr import MonomerGroup
from ..rdutils.bonding.portlib import get_num_ports


def estimate_n_atoms_linear(monomers : MonomerGroup, n_monomers : int) -> int:
    '''Given a set of monomers and the desired degree of polymerization, estimate the length of the resulting chain
    !NOTE! : As-implemented, only works for linear homopolymers and block copolymers with equal an distribution of monomers'''
    # TOSELF : omitted logging for now, as it gets repeated on EVERY cycle in when called estimate_n_monomers_supremum()
    num_mono = monomers.n_monomers
    mono_term    = np.zeros(num_mono, dtype=bool) # terminality of each monomer (i.e. whether or not it is a term group)
    mono_multip  = np.zeros(num_mono, dtype=int) # multiplicity of each polymer (i.e. how many times is occurs in a chain)
    mono_contrib = np.zeros(num_mono, dtype=int) # contribution of each monomer (i.e. how many atoms does it add to the chain)

    for i, (resname, monomer) in enumerate(monomers.iter_rdmols()):
        num_atoms = monomer.GetNumAtoms()
        num_ports = get_num_ports(monomer)
        is_term = MonomerGroup.is_terminal(monomer)

        mono_term[i] = is_term
        mono_multip[i] = is_term # temporarily set middle monomer contribution to 0
        mono_contrib[i] = num_atoms - num_ports

    num_term = sum(mono_term)
    num_mid  = num_mono - num_term # assumed that all monomers are either terminal or not
    mono_multip[~mono_term] = (n_monomers - num_term) / num_mid # naive assumption that all middle monomers contribute rest of chain equally (for homopolymers, this is always true)

    N = mono_contrib @ mono_multip # compute dot product to yield final count
    
    return N

def estimate_n_monomers_infimum(monomers : MonomerGroup, n_atoms_max : int, n_monomers_min : int=3) -> int:
    '''
    For a given collection of monomer fragments, returns the largest number of monomers which guarantees that
    a polymer chain made up of those monomers will have no more than the specified maximum number of atoms
    '''
    n_atoms_base = estimate_n_atoms_linear(monomers, n_monomers_min)
    if n_atoms_base > n_atoms_max: # pre-check when optimization is impossible
        raise InsufficientChainLength(f'Even shortest possible chain ({n_monomers_min} monomers, with {n_atoms_base} atoms) is longer than the specified max length of {n_atoms_max} atoms')

    n_monomers = n_monomers_min 
    while estimate_n_atoms_linear(monomers, n_monomers + 1) < n_atoms_max: # check if adding 1 more monomer keeps the length below the threshold
        n_monomers += 1

    return n_monomers

def estimate_n_monomers_supremum(monomers : MonomerGroup, n_atoms_min : int, n_monomers_min : int=3) -> int: # NOTE : as currently defined, this also subsumes the case when the estimate and calculated length are exactly equal
    '''
    For a given collection of monomer fragments, returns the smallest number of monomers which guarantees that
    a polymer chain made up of those monomers will have no fewer than the specified minimum number of atoms
    '''
    return estimate_n_monomers_infimum(monomers, n_atoms_min, n_monomers_min=n_monomers_min) + 1 # by definition, a ny more monomers than the infimum guarantees the chain will surpass a given number of atoms


'''Utilities for representing, converting, and formatting amounts of time'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import ClassVar, Union, TypeAlias
from dataclasses import dataclass, field

from time import time
from string import Template
from datetime import timedelta

from .typetools.categorical import _union_member_factory


# TIME CONVERSION CONSTANTS
SECONDS_PER_INTERVAL = { # hard-coded version which omits dependency on OpenMM/other unit engine
    'year'        : 31_557_600.0,
    'day'         : 86_400.0,
    'hour'        : 3_600.0,
    'minute'      : 60.0,
    'second'      : 1.0,
    'millisecond' : 1E-3,
    'microsecond' : 1E-6,
}
SECONDS_PER_INTERVAL_ORDERED = { # arrange in descending order by magnitude of conversion factor
    unit_name : factor
        for unit_name, factor in sorted(SECONDS_PER_INTERVAL.items(), key=lambda x : x[1], reverse=True)
}


# TYPING AND CONVERSION
Timeable : TypeAlias = Union[int, float, timedelta]
istimeable = _union_member_factory(Timeable, 'Timeable')

def _convert_interval_to_seconds(interval : Timeable) -> float:
    '''Takes an object interpretable as a duration in seconds and returns a float or int corresponding to that interval (in seconds)'''
    if isinstance(interval, float):
        return interval
    elif isinstance(interval, int):
        return float(interval)
    elif isinstance(interval, timedelta):
        return interval.total_seconds()
    # elif isinstance(interval, Quantity): # deprecated to avoid OpenMM requirement; may reintroduce standard unit engine for polymerist has been decided
    #     if not interval.unit.is_compatible(second):
    #         raise ValueError('Quantity must have units dimensions of time to be interpreted as an interval')
    #     return interval.in_units_of(second)._value
    else:
        raise TypeError(f'Unsupported type "{interval.__class__.__name__}" for time interval breakdown')
    

# REPRESENTING DURATIONS
class TimeTemplate(Template):
    '''Like a string Template, but which uses a percent to indicate fields (much like a date formatter)'''
    delimiter : str = '%'

@dataclass
class Duration:
    '''For representing, converting, and formatting a length of time'''
    year        : int = field(default_factory=int)
    day         : int = field(default_factory=int)
    hour        : int = field(default_factory=int)
    minute      : int = field(default_factory=int)
    second      : int = field(default_factory=int)
    millisecond : int = field(default_factory=int)
    microsecond : int = field(default_factory=int)

    # formatting constants
    _FMT_ALIASES : ClassVar[dict[str, str]] = { # hard-coded aliases for units of time for string formatting
        'year'        : 'Y',
        'day'         : 'D',
        'hour'        : 'H',
        'minute'      : 'M',
        'second'      : 'S',
        'millisecond' : 's',
        'microsecond' : 'f',

    }

    _FMT_DIGITS : ClassVar[dict[str, str]] = { # hard-coded numbers of digits for formatting
        'year'        : 3,
        'day'         : 2,
        'hour'        : 2,
        'minute'      : 2,
        'second'      : 2,
        'millisecond' : 3,
        'microsecond' : 3,

    }

    # interconversion
    @classmethod
    def from_seconds(cls, interval : Timeable) -> dict[str, int]:
        '''Takes an object interpretable as a duration in seconds and returns the breakdown by years, hours, minutes, seconds, and milliseconds'''
        time_remain = _convert_interval_to_seconds(interval)

        breakdown = {}
        for unit_name, factor in SECONDS_PER_INTERVAL_ORDERED.items():
            unit_val, time_remain = divmod(time_remain, factor) 
            breakdown[unit_name] = round(unit_val) # by definition, the quotient part must be an integer, so no precision is lost here

        return cls(**breakdown)
    
    def to_seconds(self) -> float:
        '''Convert the stored duration into a number of seconds'''
        return sum(
            unit_val * SECONDS_PER_INTERVAL[unit]
                for unit, unit_val in self.__dict__.items()
        )
    
    @property
    def total_seconds(self) -> float:
        '''conversion-to-seconds alias for convenience'''
        return self.to_seconds()
    
    # strftime-like formatting
    @property
    def _fmt_dict(self) -> dict[str, str]:
        '''Convert stored times into string format more amenable to string interpolation'''
        return {
            self._FMT_ALIASES[unit] : f'{unit_val:0{self._FMT_DIGITS[unit]}d}'
                for unit, unit_val in self.__dict__.items()
        }
    
    def format(self, fmt_str : str) -> str: # TODO : expand to cover all strftime functionality in the future
        '''Format stored duration similar to datetime.strftime (https://docs.python.org/3/library/datetime.html#datetime.datetime.strftime)'''
        template = TimeTemplate(fmt_str)
        return template.substitute(**self._fmt_dict)
    fmt = format # alias for convenience

class Timer:
    '''Simple context manager for measuring how long something takes'''
    def __init__(self) -> None:
        self.start_time = time()
        self.time_taken = None

    def __enter__(self) -> 'Timer':
        return self

    def __exit__(self, type, value, traceback) -> bool:
        self.time_taken = time() - self.start_time
        return True


'''For representing, generating, and modifying information about groups of monomer fragments'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from .repr import MonomerGroup # make monomer representation available at the module level
from .fragments import (
    PE_FRAGMENTS,
    MPD_TMC_FRAGMENTS,
    PEG_PLGA_FRAGMENTS,
)


'''Unit tests for filetree operations'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import pytest
from pathlib import Path
from anytree.iterators import PreOrderIter, LevelOrderGroupIter

from polymerist.genutils.importutils.pkginspect import get_dir_path_within_package
from polymerist.tests import data as testdata

from polymerist.genutils.fileutils.filetree import path_tree, dir_tree


DUMMY_DIR_INFO : dict[int, bool] = { # expected depth and dir status of test directory
    'dummy_dir' : (0, True),
    'subdir1'   : (1, True),
    'bar.txt'   : (1, False),
    'foo.dat'   : (1, False),
    'subdir2'   : (2, True),
    'spam.dat'  : (2, False),
    'baz.txt'   : (3, False),
}

@pytest.fixture
def dummy_dir_path() -> Path:
    return get_dir_path_within_package('dummy_dir', testdata)

@pytest.mark.parametrize('depth', [0, 1, 2, 3, 4])
def test_path_tree_depth(dummy_dir_path, depth : int) -> None:
    '''Test that max_depth restrictions are correctly applied'''
    tree = path_tree(dummy_dir_path, max_depth=depth)
    assert tree.height <= depth

@pytest.mark.parametrize('depth', [0, 1, 2, 3, 4])
def test_path_tree_output(dummy_dir_path, depth : int) -> None:
    '''Test that file names and path/dir status is correctly recorded'''
    tree = path_tree(dummy_dir_path, max_depth=depth)
    for node_group in LevelOrderGroupIter(tree):
        for node in node_group:
            expected = DUMMY_DIR_INFO[node.name]
            encountered = (node.depth, node.ppath.is_dir())
            assert(expected == encountered)

def test_path_tree_exclude(dummy_dir_path) -> None:
    '''Check that path_tree correctly excludes paths by filter conditions'''
    tree = path_tree(dummy_dir_path, max_depth=None, exclude=lambda path : path.suffix == '.dat')
    for node in PreOrderIter(tree):
        assert node.ppath.suffix != '.dat'

def test_dir_tree(dummy_dir_path) -> None:
    '''Test that dir_tree correctly filters down to just directories'''
    tree = dir_tree(dummy_dir_path, max_depth=None)
    for node in PreOrderIter(tree):
        assert node.dir.is_dir() # note the necessary change of attribute from "ppath" to "dir"

def test_dir_tree_exclude(dummy_dir_path) -> None:
    '''Test that dir_tree correctly excludes directories by filter conditions'''
    tree = dir_tree(dummy_dir_path, max_depth=None, exclude=lambda dir : dir.name.startswith('subdir'))
    for node in PreOrderIter(tree):
        assert not node.name.startswith('subdir') # note the necessary change of attribute from "ppath" to "dir"


'''For checking dimensionality and presence of units'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Any, Union, TypeVar
T = TypeVar('T')

from numpy import ndarray
from openmm.unit import (
    Unit as OpenMMUnit, 
    Quantity as OpenMMQuantity,
    length_dimension,
)
OpenMMUnitLike = Union[OpenMMUnit, OpenMMQuantity] # TODO: add union type checkers

from pint import ( # this is also the base classes for all OpenFF-style units
    Unit as PintUnit,
    Quantity as PintQuantity, 
)
PintUnitLike = Union[PintUnit, PintQuantity] # TODO: add union type checkers

Unit     = Union[PintUnit    , OpenMMUnit]
Quantity = Union[PintQuantity, OpenMMQuantity]


# CHECKING FOR AND REMOVING UNITS
class MissingUnitsError(Exception):
    pass

def hasunits(obj : Any) -> bool:
    '''Naive but effective way of checking for pint and openmm units'''
    return any(hasattr(obj, attr) for attr in ('unit', 'units')) 

def strip_units(coords : Union[T, PintQuantity, OpenMMQuantity]) -> Union[T, ndarray[Any]]:
    '''
    Sanitize coordinate tuples for cases which require unitless quantities
    Specifically needed since OpenMM and pint each have their own Quantity and Units classes
    '''
    if isinstance(coords, PintQuantity):
        return coords.magnitude # for container-like values (e.g. tuples), will always return numpy array instead (not type-safe!)
    elif isinstance(coords, OpenMMQuantity):
        return coords._value

    return coords

# CHECKING DIMENSIONALITY
def _is_volume_openmm(unitlike : OpenMMUnitLike) -> bool:
    '''Check whether an OpenMM Unit/Quantity dimensionally corresponds to a volume'''
    if isinstance(unitlike, OpenMMQuantity):
        unitlike = unitlike.unit # extract just the unit component if a Quantity is passed

    for i, (dim, exp) in enumerate(unitlike.iter_base_dimensions()):
        if i > 0:
            return False # immediate rule out if more than just one dimension is present
        
    if (dim == length_dimension) and (exp == 3.0): # if monodimensional, check that the single dimension is L^3
        return True
    return False

def _is_volume_pint(unitlike : PintUnitLike) -> bool:
    '''Check whether an Pint Unit/Quantity dimensionally corresponds to a volume'''
    return unitlike.dimensionality == '[length]**3' # "dimensionality" attr is present on both the Unit and Quantity classes in Pint

def is_volume(unitlike : Union[Unit, Quantity]) -> bool:
    '''
    Check whether a Unit or Quantity dimensionally corresponds to a volume
    Accepts both OpenMM-style and Pint-style unit-like objects
    '''
    if isinstance(unitlike, OpenMMUnitLike):
        return _is_volume_openmm(unitlike)
    elif isinstance(unitlike, PintUnitLike):
        return _is_volume_pint(unitlike)
    else:
        # raise TypeError(f'Cannot interpret object of type "{type(unitlike).__name__}" as unit-like')
        return False # strictly speaking, anything which has no notion of units cannot be a volume


'''Tools for generating and manipulating monomer connectivity graphs'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Any, ClassVar, Generator, Optional, Sequence, Union

import networkx as nx
from itertools import product as cartesian_product

from ..genutils.iteration import asiterable
from ..genutils.sequences.discernment import DISCERNMENTSolver
from ..genutils.textual.delimiters import validate_braces
from ..genutils.fileutils.jsonio.serialize import TypeSerializer


class MonomerGraph(nx.Graph):
    '''A graph representation of the connectivity of monomer fragments in a polymer topology'''
    MONOMER_NAME_ATTR : ClassVar[str]            = 'monomer_name'     # node attribute name assigned to monomer names
    FLAVOR_DICT_ATTR  : ClassVar[dict[int, int]] = 'neighbor_flavors' # node attribute name assigned to outgoing flavors for bonds to neighbor ports
    BONDTYPE_ATTR     : ClassVar[str]            = 'bondtype'         # edge attribute name assigned to bond type annotations


    # node and edge attributes
    def get_monomer_name_at_node_index(self, node_idx : int) -> Optional[str]:
        '''Recover the assigned monomer name for the node at the given index'''
        return self.nodes[node_idx].get(self.MONOMER_NAME_ATTR)
    monomer_name = get_monomer_name = get_monomer_name_at_node_idx = get_monomer_name_at_node_index

    def get_flavor_dict_at_node_index(self, node_idx : int) -> Optional[dict[int, int]]:
        '''Recover the assigned dictionary of neighbor flavors for the node at the given index'''
        return self.nodes[node_idx].get(self.FLAVOR_DICT_ATTR)
    flavor_dict = get_flavor_dict = get_flavor_dict_at_node_idx = get_flavor_dict_at_node_index


    # connectivity properties
    @property
    def num_monomers(self) -> int:
        '''Number of monomer units represented in the current polymer'''
        return self.number_of_nodes()

    @property
    def is_unbranched(self) -> bool:
        '''Whether the monomer graph represents straight chain(s) without branching'''
        return all(node_deg <= 2 for node_id, node_deg in self.degree)
    is_linear = is_unbranched

    @property
    def is_unbranched(self) -> bool:
        '''Whether the monomer graph represents straight chain(s) without branching'''
        return not self.is_unbranched
    
    @property
    def terminal_monomers(self) -> Generator[int, None, None]:
        '''Generates the indices of all nodes corresponding to terminal monomers (i.e. those wiht only one outgoing bond)'''
        for node_idx, degree in self.degree:
            if degree == 1:
                yield node_idx
    termini = leaves = terminal_monomers
    

    # topological and multi-chain properties
    @property
    def num_chains(self) -> int:
        '''The number of disconnected chains represented by the MonoGraph'''
        return nx.number_connected_components(self)

    @property
    def chains(self) -> Generator['MonomerGraph', None, None]:
        '''Generates all disconnected polymers chains in the graph sequentially'''
        for cc_nodes in nx.connected_components(self):
            yield self.subgraph(cc_nodes)

    @property
    def unique_monomer_names(self) -> set[str]:
        '''The collection of unique monomer names embedded in the graph nodes'''
        return set(nx.get_node_attributes(self, self.MONOMER_NAME_ATTR).values())


    # visualization
    def draw(self, label_monomers : bool=True, label_bonds : bool=True, **kwargs) -> None: # TODO: expand arg passing (positions, matplotlib axes, etc)
        '''Visualize graph structure with NetworkX'''
        if 'pos' not in kwargs:
            kwargs['pos'] = nx.spring_layout(self) # TODO: try other layouts to see which looks best

        monomer_labels = nx.get_node_attributes(self, self.MONOMER_NAME_ATTR) if label_monomers else None
        nx.draw(self, with_labels=True, labels=monomer_labels, **kwargs)
        
        bond_labels    = nx.get_edge_attributes(self, self.BONDTYPE_ATTR) if label_bonds else None
        nx.draw_networkx_edge_labels(self, edge_labels=bond_labels, **kwargs) # TODO: add flavor labels to drawing
    visualize = draw


    # chemical information checks
    def insert_chemical_info(self, chemical_info : dict[str, dict]) -> None:
        '''Insert SMILES, SMARTS, and atom/linker count info into nodes from minimal set of monomer info templates'''
        raise NotImplemented

    def _validate(self) -> bool:
        '''Check whether the chemical information inserted into the monomer graph is valid'''
        raise NotImplemented


    # SMILES-like in-line encodings
    ## Reading string
    @classmethod
    def from_smidge_string(cls, smidge_string : str, start_node_idx : int=0) -> 'MonomerGraph':
        '''Parse a SMIDGE ("SMILES-like Monomer Interconnectivity & Degree Graph Encoding") string and read it into a networkX Graph'''
        from .smidgelib.smidgeread import SMIDGEReader

        reader = SMIDGEReader()
        return reader.read_smidge(smidge_string, start_node_idx=start_node_idx)
    from_SMIDGE = from_smidge = from_smidge_string

    ## Writing string
    def _validate_start_node_idxs(self, start_node_idxs : Optional[Union[int, Sequence[int]]]=None) -> dict[int, int]:
        '''Check if a collection of DFS traversal start indices are valid for the current graph topology'''
        # 0) if explicitly NO ids are passed, no validation is needed
        n_chains = self.num_chains
        if start_node_idxs is None:
            return {
                i : min(chain) # assign the smallest node in each component as the starting indices
                    for i, chain in enumerate(self.chains)
            }

        # 1) check that there are enough start nodes for the present number of chains
        start_node_idxs = asiterable(start_node_idxs) # convert to iterable to handle singleton values in a unified way
        n_nodes = len(start_node_idxs)
        if n_nodes != n_chains:
            quantifier = 'few' if (n_nodes < n_chains) else 'many'
            raise ValueError(f'Provided too {quantifier} chain start indices traversal of the given graph ({n_nodes} provided for {n_chains}-chain graph)')

        # 2) check that there exists a 1:1 mapping between the provided node collection and DISTINCT connected components
        cc_order_planner = DISCERNMENTSolver(nx.connected_components(self))
        if not cc_order_planner.solution_exists(start_node_idxs, unique_bins=True):
            raise ValueError('Starting node indices provided do not uniquely correspond to distinct chains')
        else:
            cc_order = next(cc_order_planner.enumerate_choices(start_node_idxs, unique_bins=True))
            return {
                chain_idx : start_node_idx
                    for (chain_idx, start_node_idx) in zip(cc_order, start_node_idxs) # the parity of this is guaranteed by the prior length match check
            }
    
    def to_smidge_string(self, start_node_idxs :  Optional[Union[int, Sequence[int]]]=None) -> str:
        '''Convert a monomer graph into a SMIDGE ("SMILES-like Monomer Interconnectivity & Degree Graph Encoding") string'''
        from .smidgelib.smidgewrite import SMIDGEWriter

        writer = SMIDGEWriter()
        return writer.write_smidge(self, start_node_idxs=start_node_idxs)
    to_smidge = to_SMIDGE = to_smidge_string

    ## Testing string translation
    def _passes_string_conversion_tests(self) -> tuple[bool, Optional[tuple[int]]]:
        '''Developer function, tests if conversion to and from graph strings preserves the graph topology invariant to the starting node
        Returns a bool of whether test passes for all possible traversal starting positions, and tuple of positions of first failure (or None if passing)'''
        for start_idxs in cartesian_product(*[chain.nodes for chain in self.chains]):
            isostr = self.to_smidge_string(start_node_idxs=start_idxs)
            isograph = self.from_smidge_string(isostr)
            if not nx.is_isomorphic(self, isograph):
                return False, start_idxs
        else:
            return True, None
MonoGraph = MonomerGraph # alias for convenience

class MonomerGraphSerializer(TypeSerializer, python_type=MonomerGraph):
    '''JSON serializer for storing MonomerGraphs as SMIDGE strings '''
    @staticmethod
    def encode(python_obj : MonomerGraph) -> str:
        return python_obj.to_smidge_string()

    @staticmethod
    def decode(json_obj : str) -> MonomerGraph:
        return MonomerGraph.from_smidge(json_obj)


'''Decorators for modifying classes'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Callable, Iterable, Optional, TypeVar, Union
C = TypeVar('C')


def generate_repr(cls : Optional[C]=None, disp_attrs : Optional[Iterable[str]]=None, lookup_attr : Optional[str]=None) -> Union[C, Callable[[C], C]]:
    '''
    Class decorator for auto-generating __repr__ methods

    By default (i.e. with no arguments), generated repr simply returns the name of the class
    If collection of "disp_attrs" is provided, will display the values of <disp_attrs> for the object instance being represented in series
    If "disp_attrs" is NOT provided but "lookup_attr" is, disp_attrs will be looked up from the modified parent class
    '''
    if disp_attrs is None: 
        disp_attrs = [] # set to empty list to avoid mutable default
    
    def class_decorator(cls : C) -> C:
        '''The actual (argument-free) class decorator'''
        nonlocal disp_attrs # avoids multiple scope issues in disparate use cases (refers to the variable in the outermost scope)
        if not disp_attrs and lookup_attr: # only use lookup if one is explicitly provided and no display attributes are provided
            assert(hasattr(cls, lookup_attr))
            disp_attrs = getattr(cls, lookup_attr) # if a lookup attribute is provided, lookup the attribute names within the class being modified

        def _repr_generic(self) -> str:
            attr_str = ', '.join(f'{attr}={getattr(self, attr)}' for attr in disp_attrs)
            return f'{cls.__name__}({attr_str})'
        setattr(cls, '__repr__', _repr_generic)

        return cls
    
    if cls is None: # null case (i.e. call without parens), return factory call
        return class_decorator
    return class_decorator(cls) # return literal class decorator call
        
def register_subclasses(cls : Optional[C]=None, key_attr : str='__name__', reg_attr : str='subclass_registry') -> Union[C, Callable[[C], C]]:
    '''
    Parametric class decorator for automatically generating a registry of subclasses of a target class
    Binds registry to the "registry" class property in the target class
    
    Subclasses are keyed by lookup of a target attribute <key_attr> in the child classes (by default just the name of the subclass),
    while the resulting registry class property is bound to the <reg_attr> attribute of the parent class
    '''
    def class_decorator(cls : C) -> C:
        '''The actual (argument-free) class decorator'''
        @classmethod # method should be accessible class-wide
        @property    # make property to allow for dynamic subclassing (generated at runtime, not compile time)
        def _registry(cls : C) -> dict[str, C]:
            return { # Keep a registry of all charger implementations for convenience
                getattr(subclass, key_attr) : subclass
                    for subclass in cls.__subclasses__()
            }
        setattr(cls, reg_attr, _registry) # bind registry class property to target class. TODO : check for registry already present in class
        
        return cls # return back the modified class
    
    if cls is None: # null case (i.e. call without parens), return factory call
        return class_decorator
    return class_decorator(cls) # return literal class decorator call

# NOTE: "klass" is needed to distinguish between the class modified by this decorator and the classmethod arg when calling super()
# "klass" here is the parent, while "cls" is the child
def register_abstract_class_attrs(*attr_names : list[str]) -> Callable[[C], C]: # TODO: add mechanism for typehinting
    '''Register a list of string attribute names as abstract class attributes, 
    which MUST be implemented by child classes of the wrapped class'''
    def class_decorator(klass : C) -> C:
        '''The actual (argument-free) class decorator'''
        def __wrapped_init_subclass__(cls : C, **kwargs) -> None:
            '''Wrapper for subclass definition which actually enforces that all named attributes are set'''
            for attr_name in attr_names:
                passed_attr_value = kwargs.pop(attr_name, NotImplemented) # want this removed from kwargs before passing to super, regardless of whether already set in child 
                attr_val_on_child = getattr(cls, attr_name, NotImplemented) # check if this has been set in the child in code
                
                if attr_val_on_child is NotImplemented:         # if the value has not been set in code...
                    if passed_attr_value is not NotImplemented: # ...fall back to value passed into class definition, if it exists...
                        setattr(cls, attr_name, passed_attr_value)
                    else:                                       # otherwise, fail and raise Exception
                        raise TypeError(f"Can't instantiate abstract class {cls.__name__} with abstract class property '{attr_name}' undefined")

            super(klass, cls).__init_subclass__(**kwargs) # this should fail if extraneous named args are passed

        klass.__init_subclass__ = classmethod(__wrapped_init_subclass__)
        return klass

    return class_decorator # no need for application check here, since the parameterized decorator doesn't take a class to be modified